"""
Miscellaneous utilities and scripts of fractopo.
"""
from itertools import count
from typing import List, Tuple, Union

import geopandas as gpd
from shapely.geometry import LineString, Point

from fractopo.general import (
    compare_unit_vector_orientation,
    create_unit_vector,
    geom_bounds,
    get_trace_endpoints,
    pygeos_spatial_index,
    safe_buffer,
    spatial_index_intersection,
)


class LineMerge:

    """
    Merge lines conditionally.
    """

    @staticmethod
    def conditional_linemerge(
        first: LineString, second: LineString, tolerance: float, buffer_value: float
    ) -> Union[None, LineString]:
        """
        Conditionally merge two LineStrings (first and second).

        Merge occurs if 1) their endpoints are within buffer_value of each
        other and 2) their total orientations are within tolerance (degrees) of
        each other.

        Merges by joining their coordinates. The endpoint
        (that is within buffer_value of endpoint of first) of the second
        LineString is trimmed from the resulting coordinates.

        E.g. with merging:

        >>> first = LineString([(0, 0), (0, 2)])
        >>> second = LineString([(0, 2.001), (0, 4)])
        >>> tolerance = 5
        >>> buffer_value = 0.01
        >>> LineMerge.conditional_linemerge(first, second, tolerance, buffer_value).wkt
        'LINESTRING (0 0, 0 2, 0 4)'

        Without merging:

        >>> first = LineString([(0, 0), (0, 2)])
        >>> second = LineString([(0, 2.1), (0, 4)])
        >>> tolerance = 5
        >>> buffer_value = 0.01
        >>> LineMerge.conditional_linemerge(
        ...     first, second, tolerance, buffer_value
        ... ) is None
        True

        """
        assert isinstance(first, LineString) and isinstance(second, LineString)
        # Get trace endpoints
        first_start, first_end = get_trace_endpoints(first)
        second_start, second_end = get_trace_endpoints(second)

        # Get unit vectors from endpoints
        first_unit_vector = create_unit_vector(first_start, first_end)
        second_unit_vector = create_unit_vector(second_start, second_end)

        # Check if unit vectors are close in orientation
        are_close = compare_unit_vector_orientation(
            first_unit_vector, second_unit_vector, threshold_angle=tolerance
        )
        are_close_reverse = compare_unit_vector_orientation(
            -first_unit_vector, second_unit_vector, threshold_angle=tolerance
        )
        # Get coordinates
        first_coords_list = list(first.coords)
        second_coords_list = list(second.coords)
        if (
            first_end.buffer(buffer_value).intersects(second_start.buffer(buffer_value))
            and are_close
        ):
            # First ends in the start of second -> Sequence of coords is correct
            # Do not include first coordinate of second in new
            new_coords = first_coords_list + second_coords_list[1:]
        elif (
            first_end.buffer(buffer_value).intersects(second_end.buffer(buffer_value))
            and are_close_reverse
        ):
            # First ends in second end
            new_coords = first_coords_list + list(reversed(second_coords_list))[1:]
        elif (
            first_start.buffer(buffer_value).intersects(
                second_start.buffer(buffer_value)
            )
            and are_close_reverse
        ):
            # First starts from the same as second
            new_coords = list(reversed(second_coords_list))[:-1] + first_coords_list
        elif (
            first_start.buffer(buffer_value).intersects(second_end.buffer(buffer_value))
            and are_close
        ):
            # First starts from end of second
            new_coords = second_coords_list[:-1] + first_coords_list
        else:
            return None
        return LineString(new_coords)

    @staticmethod
    def conditional_linemerge_collection(
        traces: Union[gpd.GeoDataFrame, gpd.GeoSeries],
        tolerance: float,
        buffer_value: float,
    ) -> Tuple[List[LineString], List[int]]:
        """
        Conditionally linemerge within a collection of LineStrings.

        Returns the linemerged traces and the idxs of traces that were
        linemerged.

        E.g.


        >>> first = LineString([(0, 0), (0, 2)])
        >>> second = LineString([(0, 2.001), (0, 4)])
        >>> traces = gpd.GeoSeries([first, second])
        >>> tolerance = 5
        >>> buffer_value = 0.01
        >>> new_traces, idx = LineMerge.conditional_linemerge_collection(
        ...     traces, tolerance, buffer_value
        ... )
        >>> [trace.wkt for trace in new_traces], idx
        (['LINESTRING (0 0, 0 2, 0 4)'], [0, 1])

        """
        spatial_index = pygeos_spatial_index(traces)

        new_traces = []
        modified_idx = []
        for i, trace in enumerate(traces.geometry):
            assert isinstance(trace, LineString)
            trace_candidates_idx: List[int] = list(
                spatial_index.intersection(
                    geom_bounds(safe_buffer(trace, buffer_value * 2))
                )
            )
            trace_candidates_idx.remove(i)
            if len(trace_candidates_idx) == 0 or i in modified_idx:
                continue
            for idx in trace_candidates_idx:
                if idx in modified_idx or i in modified_idx:
                    continue
                trace_candidate = traces.geometry.iloc[idx]

                merged = LineMerge.conditional_linemerge(
                    trace,
                    trace_candidate,
                    tolerance=tolerance,
                    buffer_value=buffer_value,
                )
                if merged is not None:
                    new_traces.append(merged)
                    modified_idx.append(i)
                    modified_idx.append(idx)
                    break

        return new_traces, modified_idx

    @staticmethod
    def run_loop(
        traces: gpd.GeoDataFrame, tolerance: float, buffer_value: float
    ) -> gpd.GeoDataFrame:
        """
        Run multiple conditional linemerge iterations for GeoDataFrame.

        This is the main entrypoint.

        GeoDataFrame should contain LineStrings.

        E.g.

        >>> first = LineString([(0, 0), (0, 2)])
        >>> second = LineString([(0, 2.001), (0, 4)])
        >>> traces = gpd.GeoDataFrame(geometry=[first, second])
        >>> tolerance = 5
        >>> buffer_value = 0.01
        >>> LineMerge.run_loop(traces, tolerance, buffer_value)
                                                    geometry
        0  LINESTRING (0.00000 0.00000, 0.00000 2.00000, ...

        """
        loop_count = count()
        while True:
            new_traces, modified_idx = LineMerge.conditional_linemerge_collection(
                traces, tolerance=tolerance, buffer_value=buffer_value
            )
            if len(modified_idx) == 0:
                return traces
            traces = LineMerge.integrate_replacements(traces, new_traces, modified_idx)
            if next(loop_count) > 100:
                print("loop 100")
                return traces

    @staticmethod
    def integrate_replacements(
        traces: gpd.GeoDataFrame, new_traces: List[LineString], modified_idx: List[int]
    ) -> gpd.GeoDataFrame:
        """
        Add linemerged and remove the parts that were linemerged.

        E.g.


        >>> first = LineString([(0, 0), (0, 2)])
        >>> second = LineString([(0, 2.001), (0, 4)])
        >>> traces = gpd.GeoDataFrame(geometry=[first, second])
        >>> new_traces = [LineString([(0, 0), (0, 2), (0, 4)])]
        >>> modified_idx = [0, 1]
        >>> LineMerge.integrate_replacements(traces, new_traces, modified_idx)
                                                    geometry
        0  LINESTRING (0.00000 0.00000, 0.00000 2.00000, ...

        """
        unmod_traces = [
            trace
            for idx, trace in enumerate(traces.geometry)
            if idx not in modified_idx
        ]
        all_traces = unmod_traces + new_traces
        gdf = gpd.GeoDataFrame(geometry=all_traces)
        if traces.crs is not None:
            gdf = gdf.set_crs(traces.crs)
        return gdf


def remove_identical_sindex(
    geosrs: gpd.GeoSeries, snap_threshold: float
) -> gpd.GeoSeries:
    """
    Remove stacked nodes by using a search buffer the size of snap_threshold.
    """
    geosrs_reset = geosrs.reset_index(inplace=False, drop=True)
    assert isinstance(geosrs_reset, gpd.GeoSeries)
    geosrs = geosrs_reset
    spatial_index = geosrs.sindex
    identical_idxs = []
    point: Point
    for idx, point in enumerate(geosrs.geometry.values):
        if idx in identical_idxs:
            continue
        # point = point.buffer(snap_threshold) if snap_threshold != 0 else point
        p_candidate_idxs = (
            # list(spatial_index.intersection(point.buffer(snap_threshold).bounds))
            spatial_index_intersection(
                spatial_index=spatial_index,
                coordinates=geom_bounds(safe_buffer(geom=point, radius=snap_threshold)),
            )
            if snap_threshold != 0
            else list(spatial_index.intersection(point.coords[0]))
        )
        p_candidate_idxs.remove(idx)
        p_candidates = geosrs.iloc[p_candidate_idxs]
        inter = p_candidates.distance(point) < snap_threshold
        colliding = inter.loc[inter]
        if len(colliding) > 0:
            index_to_list = colliding.index.to_list()
            assert len(index_to_list) > 0
            assert all(isinstance(i, int) for i in index_to_list)
            identical_idxs.extend(index_to_list)
    geosrs_dropped = geosrs.drop(identical_idxs)
    assert isinstance(geosrs_dropped, gpd.GeoSeries)
    return geosrs_dropped



"""
Functions for plotting cross-cutting and abutting relationships.
"""

import logging
from itertools import chain, combinations
from typing import Dict, List, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.ticker import MaxNLocator
from shapely import prepared
from shapely.geometry import LineString, MultiPoint, Point

from fractopo.general import (
    X_node,
    Y_node,
    get_trace_endpoints,
    prepare_geometry_traces,
)

log = logging.getLogger(__name__)


def determine_crosscut_abutting_relationships(
    trace_series: gpd.GeoSeries,
    node_series: gpd.GeoSeries,
    node_types: np.ndarray,
    set_array: np.ndarray,
    set_names: Tuple[str, ...],
    buffer_value: float,
    label: str,
) -> pd.DataFrame:
    """
    Determine cross-cutting and abutting relationships between trace sets.

    Determines relationships between all inputted sets by using spatial
    intersects between node and trace data.

    E.g.

    >>> trace_series = gpd.GeoSeries(
    ...     [LineString([(0, 0), (1, 0)]), LineString([(0, 1), (0, -1)])]
    ... )
    >>> node_series = gpd.GeoSeries(
    ...     [Point(0, 0), Point(1, 0), Point(0, 1), Point(0, -1)]
    ... )
    >>> node_types = np.array(["Y", "I", "I", "I"])
    >>> set_array = np.array(["1", "2"])
    >>> set_names = ("1", "2")
    >>> buffer_value = 0.001
    >>> label = "title"
    >>> determine_crosscut_abutting_relationships(
    ...     trace_series,
    ...     node_series,
    ...     node_types,
    ...     set_array,
    ...     set_names,
    ...     buffer_value,
    ...     label,
    ... )
        name    sets  x  y y-reverse error-count
    0  title  (1, 2)  0  1         0           0

    TODO: No within set relations.....yet... Problem?

    """
    assert len(set_array) == len(trace_series)
    assert len(node_series) == len(node_types)
    assert all(isinstance(val, LineString) for val in trace_series.geometry.values)
    # Determines xy relations and dynamically creates a dataframe as an aid for
    # plotting the relations

    relations_df = pd.DataFrame(
        columns=["name", "sets", "x", "y", "y-reverse", "error-count"]
    )

    is_xy = np.array([val in (X_node, Y_node) for val in node_types])
    node_series_xy: gpd.GeoSeries = node_series.loc[is_xy]  # type: ignore
    node_types_xy = node_types[is_xy]

    assert len(node_series_xy) == len(node_types_xy)

    if len(set_names) < 2:
        raise ValueError("Expected more than one set to be defined.")

    set_combinations = combinations(set_names, 2)
    additions = []
    for first_set, second_set in set_combinations:
        trace_series_two_sets: Tuple[gpd.GeoSeries, gpd.GeoSeries] = (
            trace_series.loc[set_array == first_set],  # type: ignore
            trace_series.loc[set_array == second_set],  # type: ignore
        )

        if any(series.shape[0] == 0 for series in trace_series_two_sets):
            log.warning("Expected first_set and second_set to both contain traces.")
            return relations_df
        set_names_two_sets = (first_set, second_set)
        # determine_nodes_intersecting_sets returns a boolean array
        # representing nodes that intersect both sets.
        intersects_both_sets = determine_nodes_intersecting_sets(
            trace_series_two_sets=trace_series_two_sets,  # type: ignore
            set_names_two_sets=set_names_two_sets,
            node_series_xy=node_series_xy,  # type: ignore
            buffer_value=buffer_value,
        )
        assert len(node_series_xy) == len(node_types_xy) == len(intersects_both_sets)
        node_series_xy_intersects = node_series_xy.loc[intersects_both_sets]
        node_types_xy_intersects = node_types_xy[intersects_both_sets]

        intersectframe = determine_intersects(
            trace_series_two_sets=trace_series_two_sets,
            node_series_xy_intersects=node_series_xy_intersects,  # type: ignore
            node_types_xy_intersects=node_types_xy_intersects,
            set_names_two_sets=set_names_two_sets,
            buffer_value=buffer_value,
        )
        intersect_series = intersectframe.groupby(["nodeclass", "sets"]).size()

        x_count = 0
        y_count = 0
        y_reverse_count = 0

        for item in list(intersect_series.items()):
            value = item[1]
            if item[0][0] == X_node:
                x_count = value
            elif item[0][0] == Y_node:
                if item[0][1] == (
                    first_set,
                    second_set,
                ):  # it's set s abutting set second_set
                    y_count = value
                elif item[0][1] == (
                    second_set,
                    first_set,
                ):  # it's set second_set abutting set s
                    y_reverse_count = value
                else:
                    raise ValueError(
                        f"item[0][1] does not equal {(first_set, second_set)}"
                        f" nor {(second_set, first_set)}\nitem[0][1]: {item[0][1]}"
                    )
            else:
                raise ValueError(
                    f'item[0][0] does not match "X" or "Y"\nitem[0][0]: {item[0][0]}'
                )

        addition = {
            "name": label,
            "sets": (first_set, second_set),
            "x": x_count,
            "y": y_count,
            "y-reverse": y_reverse_count,
            "error-count": len(intersectframe.loc[intersectframe.error]),
        }
        additions.append(addition)

    additions_df = pd.DataFrame(additions)
    relations_df = pd.concat([relations_df, additions_df], ignore_index=True)
    return relations_df


def determine_nodes_intersecting_sets(
    trace_series_two_sets: Tuple[gpd.GeoSeries, gpd.GeoSeries],
    set_names_two_sets: Tuple[str, str],
    node_series_xy: gpd.GeoSeries,
    buffer_value: float,
) -> List[bool]:
    """
    Conduct a spatial intersect between nodes and traces.

    Node GeoDataFrame contains only X- and Y-nodes and the trace GeoDataFrame
    only two sets. Returns boolean array of based on intersections.

    E.g.

    >>> traces = gpd.GeoSeries([LineString([(0, 0), (1, 1)])]), gpd.GeoSeries(
    ...     [LineString([(0, 1), (0, -1)])]
    ... )
    >>> set_names_two_sets = ("1", "2")
    >>> nodes_xy = gpd.GeoSeries([Point(0, 0), Point(1, 1), Point(0, 1), Point(0, -1)])
    >>> buffer_value = 0.001
    >>> determine_nodes_intersecting_sets(
    ...     traces, set_names_two_sets, nodes_xy, buffer_value
    ... )
    [True, False, False, False]

    """

    def _intersects_both(
        point: Point,
        prep_traces_first: prepared.PreparedGeometry,
        prep_traces_second: prepared.PreparedGeometry,
        buffer_value: float,
    ):
        return prep_traces_first.intersects(
            point.buffer(buffer_value)
        ) and prep_traces_second.intersects(point.buffer(buffer_value))

    assert len(set_names_two_sets) == 2
    trace_series_first_set = trace_series_two_sets[0]
    trace_series_second_set = trace_series_two_sets[1]

    prep_traces_first = prepare_geometry_traces(trace_series_first_set)
    prep_traces_second = prepare_geometry_traces(trace_series_second_set)
    intersects_both_sets = [
        _intersects_both(
            point, prep_traces_first, prep_traces_second, buffer_value=buffer_value
        )
        for point in node_series_xy.geometry.values
    ]
    return intersects_both_sets


def determine_intersects(
    trace_series_two_sets: Tuple[gpd.GeoSeries, gpd.GeoSeries],
    set_names_two_sets: Tuple[str, str],
    node_series_xy_intersects: gpd.GeoSeries,
    node_types_xy_intersects: np.ndarray,
    buffer_value: float,
) -> pd.DataFrame:
    """
    Determine how abutments and crosscuts occur between two sets.

    E.g.

    >>> traces = gpd.GeoSeries([LineString([(0, 0), (1, 1)])]), gpd.GeoSeries(
    ...     [LineString([(0, 1), (0, -1)])]
    ... )
    >>> set_names_two_sets = ("1", "2")
    >>> node_series_xy_intersects = gpd.GeoSeries([Point(0, 0)])
    >>> node_types_xy_intersects = np.array(["Y"])
    >>> buffer_value = 0.001
    >>> determine_intersects(
    ...     traces,
    ...     set_names_two_sets,
    ...     node_series_xy_intersects,
    ...     node_types_xy_intersects,
    ...     buffer_value,
    ... )
              node nodeclass    sets  error
    0  POINT (0 0)         Y  (1, 2)  False

    """
    # TODO: No DataFrames -> Refactor to something else
    # TODO: Refactor intersect logic (but it works for now)
    intersectframe = pd.DataFrame(columns=["node", "nodeclass", "sets", "error"])
    if len(node_series_xy_intersects) == 0:
        # No intersections between sets
        log.debug("No intersections between sets.")
        return intersectframe

    first_set, second_set = (
        set_names_two_sets[0],
        set_names_two_sets[1],
    )  # sets for comparison

    trace_series_first_set = trace_series_two_sets[0]
    trace_series_second_set = trace_series_two_sets[1]

    first_set_prep = prepare_geometry_traces(trace_series_first_set)
    second_set_prep = prepare_geometry_traces(trace_series_second_set)
    # Creates a rtree from all start- and endpoints of set 1
    # Used in deducting in which set a trace abuts (Y-node)
    first_set_points = list(
        chain(*list(trace_series_first_set.geometry.apply(get_trace_endpoints).values))
    )
    assert all(isinstance(p, Point) for p in first_set_points)
    # first_setpointtree = STRtree(first_set_points)
    first_setpointtree = prepared.prep(MultiPoint(first_set_points))
    node: Point
    node_class: str
    additions = []
    for node, node_class in zip(node_series_xy_intersects, node_types_xy_intersects):
        assert isinstance(node, Point)
        assert isinstance(node_class, str)
        # for idx, row in intersecting_nodes_frame.iterrows():
        # node = row.geometry
        # c = row.c

        # Checks if node intersects set 1 traces.
        l1 = first_set_prep.intersects(node.buffer(buffer_value))
        # Checks if node intersects set 2 traces.
        l2 = second_set_prep.intersects(node.buffer(buffer_value))

        if (l1 is False) and (l2 is False):  # DEBUGGING
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )

        try:
            addition = determine_intersect(
                node=node,
                node_class=node_class,
                l1=l1,
                l2=l2,
                first_set=first_set,
                second_set=second_set,
                first_setpointtree=first_setpointtree,
                buffer_value=buffer_value,
            )
        except ValueError:
            # NO RELATIONS FOR NODE IS GIVEN AS ERROR == TRUE (ERROR).
            # addition gets overwritten if there are no errors
            addition = {
                "node": node,
                "nodeclass": node_class,
                "sets": set_names_two_sets,
                "error": True,
            }  # DEBUGGING
        additions.append(addition)

    additions_df = pd.DataFrame(additions)
    intersectframe = pd.concat([intersectframe, additions_df], ignore_index=True)
    # intersectframe = intersectframe.append(
    #     addition, ignore_index=True
    # )  # Append frame with result

    return intersectframe


def determine_intersect(
    node: Point,
    node_class: str,
    l1: bool,
    l2: bool,
    first_set: str,
    second_set: str,
    first_setpointtree: prepared.PreparedGeometry,
    buffer_value: float,
) -> Dict[str, Union[Point, str, Tuple[str, str], bool]]:
    """
    Determine what intersection the node represents.

    TODO: R0912: Too many branches.
    """
    if node_class == "X":
        if l1 and l2:  # It's an x-node between sets
            sets = (first_set, second_set)
            addition = {
                "node": node,
                "nodeclass": node_class,
                "sets": sets,
                "error": False,
            }

        elif l1 and not l2:  # It's an x-node inside set 1
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )
            # sets = (first_set, first_set)
            # addition = {'node': node, 'nodeclass': c, 'sets': sets}

        elif not l1 and l2:  # It's an x-node inside set 2
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )
            # sets = (second_set, second_set)
            # addition = {'node': node, 'nodeclass': c, 'sets': sets}
        else:
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )

    # ALL Y NODE RELATIONS
    elif node_class == "Y":
        if (l1 is True) and (l2 is True):  # It's an y-node between sets
            # p1 == length of list of nodes from first_set traces
            # that intersect with X- or Y-node
            # p1 = len(first_setpointtree.query(node.buffer(buffer_value)))
            p1 = first_setpointtree.intersects(node.buffer(buffer_value))
            if p1:  # set 1 ends in set 2
                sets = (first_set, second_set)
            else:  # set 2 ends in set 1
                sets = (second_set, first_set)
            addition = {
                "node": node,
                "nodeclass": node_class,
                "sets": sets,
                "error": False,
            }

        elif (l1 is True) and (l2 is False):  # It's a y-node inside set 1
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )
            # sets = (first_set, first_set)
            # addition = {'node': node, 'nodeclass': c, 'sets': sets}

        elif (l1 is False) and (l2 is True):  # It's a y-node inside set 2
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )
            # sets = (second_set, second_set)
            # addition = {'node': node, 'nodeclass': c, 'sets': sets}
        else:
            raise ValueError(
                f"Node {node} does not intersect both sets"
                f" {first_set} and {second_set}\n l1 is {l1} and l2 is {l2}"
            )
    else:
        raise ValueError(f"Node {node} neither X or Y")

    return addition


def plot_crosscut_abutting_relationships_plot(
    relations_df: pd.DataFrame, set_array: np.ndarray, set_names: Tuple[str, ...]
) -> Tuple[List[Figure], List[np.ndarray]]:
    """
    Plot cross-cutting and abutting relationships.
    """
    # relations_df = pd.DataFrame(
    #     columns=["name", "sets", "x", "y", "y-reverse", "error-count"]
    # )
    # SUBPLOTS, FIGURE SETUP
    # set_column = "set"
    cols = relations_df.shape[0]
    if cols == 0:
        log.warning("Expected relations_df to have rows. Returning empty lists.")
        return [], []
    if cols == 2:
        cols = 1
    width = 12 / 3 * cols
    height = (width / cols) * 0.75
    names = set(relations_df.name.tolist())
    figs: List[Figure] = []
    fig_axes = []
    with plt.style.context("default"):
        for name in names:
            relations_df_with_name = relations_df.loc[relations_df.name == name]
            set_counts = []
            for set_name in set_names:
                set_counts.append(sum(set_array == set_name))

            fig, axes = plt.subplots(ncols=cols, nrows=1, figsize=(width, height))
            if not isinstance(axes, np.ndarray):
                axes = np.array([axes])

            prop_title = dict(
                boxstyle="square", facecolor="linen", alpha=1, linewidth=2
            )

            fig.suptitle(
                f"   {name}   ",
                x=0.19,
                y=1.0,
                fontsize=20,
                fontweight="bold",
                fontfamily="DejaVu Sans",
                va="center",
                bbox=prop_title,
            )

            for ax, idx_row in zip(axes, relations_df_with_name.iterrows()):
                row = idx_row[1]
                # TODO: More colors? change brightness or some other parameter?
                bars = ax.bar(
                    x=[0.3, 0.55, 0.65],
                    height=[row["x"], row["y"], row["y-reverse"]],
                    width=0.1,
                    color=["darkgrey", "darkolivegreen", "darkseagreen"],
                    linewidth=1,
                    edgecolor="black",
                    alpha=0.95,
                    zorder=10,
                )

                ax.legend(
                    bars,
                    (
                        f"Sets {row.sets[0]} and {row.sets[1]} cross-cut",
                        f"Set {row.sets[0]} abuts set {row.sets[1]}",
                        f"Set {row.sets[1]} abuts set {row.sets[0]}",
                    ),
                    framealpha=0.6,
                    loc="upper center",
                    edgecolor="black",
                    prop={"family": "DejaVu Sans"},
                )

                ax.set_ylim(0, 1.6 * max([row["x"], row["y"], row["y-reverse"]]))

                ax.grid(zorder=-10, color="black", alpha=0.5)

                xticks = [0.3, 0.6]
                xticklabels = ["X", "Y"]
                ax.xaxis.set_ticks(
                    ticks=xticks,
                    labels=xticklabels,
                )

                xticklabels_texts = ax.get_xticklabels()

                for xtick in xticklabels_texts:
                    xtick.set_fontweight("bold")
                    xtick.set_fontsize(12)

                ax.set_xlabel(
                    "Node type",
                    fontweight="bold",
                    fontsize=13,
                    fontstyle="italic",
                    fontfamily="DejaVu Sans",
                )
                ax.set_ylabel(
                    "Node count",
                    fontweight="bold",
                    fontsize=13,
                    fontstyle="italic",
                    fontfamily="DejaVu Sans",
                )

                # Set y ticks as integers
                ax.yaxis.set_major_locator(MaxNLocator(integer=True))

                plt.subplots_adjust(wspace=0.3)

                if ax == axes[-1]:
                    text = ""
                    prop = dict(boxstyle="square", facecolor="linen", alpha=1, pad=0.45)
                    for set_label, set_len in zip(set_names, set_counts):
                        text += f"Set {set_label} trace count: {set_len}"
                        if not set_label == set_names[-1]:
                            text += "\n"
                    ax.text(
                        1.1,
                        0.5,
                        text,
                        rotation=90,
                        transform=ax.transAxes,
                        va="center",
                        bbox=prop,
                        fontfamily="DejaVu Sans",
                    )
            figs.append(fig)
            fig_axes.append(axes)
    return figs, fig_axes



"""
Analyse and plot trace map data with Network.
"""
import logging
from dataclasses import dataclass, field
from functools import wraps
from pathlib import Path
from textwrap import dedent
from typing import Callable, Dict, List, Optional, Tuple, Union

import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import powerlaw
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.projections import PolarAxes
from mpl_toolkits.axes_grid1 import make_axes_locatable
from shapely.geometry import MultiPoint, MultiPolygon, Point, Polygon
from ternary.ternary_axes_subplot import TernaryAxesSubplot

from fractopo.analysis.anisotropy import determine_anisotropy_sum, plot_anisotropy_plot
from fractopo.analysis.azimuth import AzimuthBins
from fractopo.analysis.contour_grid import run_grid_sampling
from fractopo.analysis.length_distributions import LengthDistribution
from fractopo.analysis.line_data import LineData
from fractopo.analysis.parameters import (
    branches_intersect_boundary,
    convert_counts,
    determine_branch_type_counts,
    determine_node_type_counts,
    determine_topology_parameters,
    plot_parameters_plot,
    plot_ternary_plot,
)
from fractopo.analysis.relationships import (
    determine_crosscut_abutting_relationships,
    plot_crosscut_abutting_relationships_plot,
)
from fractopo.branches_and_nodes import branches_and_nodes
from fractopo.general import (
    CENSORING,
    CLASS_COLUMN,
    CONNECTION_COLUMN,
    NAME,
    RADIUS,
    RELATIVE_CENSORING,
    REPRESENTATIVE_POINT,
    EE_branch,
    Number,
    Param,
    SetRangeTuple,
    assign_branch_and_node_colors,
    bool_arrays_sum,
    calc_circle_radius,
    check_for_z_coordinates,
    crop_to_target_areas,
    determine_boundary_intersecting_lines,
    focus_plot_to_bounds,
    numpy_to_python_type,
    pygeos_spatial_index,
    raise_determination_error,
    remove_z_coordinates_from_geodata,
    sanitize_name,
    save_fig,
    spatial_index_intersection,
    total_bounds,
    write_geodata,
    write_geodataframe,
)

log = logging.getLogger(__name__)


def requires_topology(func: Callable) -> Callable:
    """
    Wrap methods that require determined topology.

    Raises an error if trying to call them without determined topology.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        if "network" in kwargs:
            network = kwargs.pop("network")
        else:
            network = args[0]
            args = args[1:] if len(args) > 1 else []
        if not network.topology_determined:
            raise_determination_error(func.__name__)
        return func(network, *args, **kwargs)

    return wrapper


@dataclass
class Network:

    """
    Trace network.

    Consists of at its simplest of validated traces and a target area that
    delineates the traces i.e., only ``trace_gdf`` and ``area_gdf`` parameters
    are required to run the network analysis but results might not be correct
    or match your expectations e.g., traces are truncated to target area
    by default.

    :param trace_gdf: ``GeoDataFrame`` containing trace data
        i.e. ``shapely.geometry.LineString`` geometries.
    :param area_gdf: ``GeoDataFrame`` containing
        target area data i.e. ``(Multi)Polygon's``.
    :param name: Name the Network.
    :param determine_branches_nodes: Whether to determine branches and nodes.
    :param snap_threshold: The snapping distance threshold to identify
        snapped traces.
    :param truncate_traces: Whether to crop the traces at the target area
        boundary.
    :param circular_target_area: Is the target are a circle.
    :param azimuth_set_names: Names of each azimuth set.
    :param azimuth_set_ranges: Ranges of each azimuth set.
    :param trace_length_set_names: Names of each trace length set.
    :param trace_length_set_ranges: Ranges of each trace length set.
    :param branch_length_set_names: Names of each branch length set.
    :param branch_length_set_ranges: Ranges of each branch length set.
    :param branch_gdf: ``GeoDataFrame`` containing branch data.
        It is recommended to let ``fractopo.Network`` determine both
        branches and nodes instead of passing them here.
    :param node_gdf: GeoDataFrame containing node data.
        It is recommended to let ``fractopo.Network`` determine both
        branches and nodes instead of passing them here.
    :param censoring_area: Polygon(s) in ``GeoDataFrame`` that delineate(s) the
        area in which trace digitization was uncertain due to censoring caused
        by e.g. vegetation.
    :param cache_results: Whether to use joblib memoize to disk-cache computationally
        expensive results.
    """

    # Base data
    # =========
    trace_gdf: gpd.GeoDataFrame
    area_gdf: gpd.GeoDataFrame

    # Name the network for e.g. plot titles
    name: str = "Network"

    determine_branches_nodes: bool = False
    snap_threshold: float = 0.001

    # The traces can be cut to end at the boundary of the target area
    # Defaults to True
    truncate_traces: bool = True

    # Whether to apply boundary line length weighting
    # Applies to both traces and branches
    # Length of lines that intersect the boundary once are multiplied by two,
    # and double intersections with 0, non-intersecting are multiplied by one
    # (no change).
    circular_target_area: bool = False

    # Azimuth sets
    # ============
    azimuth_set_names: Tuple[str, ...] = ("1", "2", "3")
    azimuth_set_ranges: SetRangeTuple = (
        (0, 60),
        (60, 120),
        (120, 180),
    )

    # Length sets
    # ===========

    # Trace length
    trace_length_set_names: Tuple[str, ...] = ()
    trace_length_set_ranges: SetRangeTuple = ()

    # Branch length
    branch_length_set_names: Tuple[str, ...] = ()
    branch_length_set_ranges: SetRangeTuple = ()

    # Branches and nodes
    # ==================
    branch_gdf: gpd.GeoDataFrame = field(default_factory=gpd.GeoDataFrame)
    node_gdf: gpd.GeoDataFrame = field(default_factory=gpd.GeoDataFrame)

    censoring_area: gpd.GeoDataFrame = field(default_factory=gpd.GeoDataFrame)

    # Misc
    # ==================
    cache_results: bool = True
    remove_z_coordinates_from_inputs: bool = True

    # Private caching attributes
    # ==========================
    _anisotropy: Optional[Tuple[np.ndarray, np.ndarray]] = field(
        default=None, repr=False
    )
    _parameters: Optional[Dict[str, float]] = field(default=None, repr=False)
    _azimuth_set_relationships: Optional[pd.DataFrame] = field(default=None, repr=False)
    _trace_length_set_relationships: Optional[pd.DataFrame] = field(
        default=None, repr=False
    )
    _trace_intersects_target_area_boundary: Optional[np.ndarray] = field(
        default=None, repr=False
    )
    _branch_intersects_target_area_boundary: Optional[np.ndarray] = field(
        default=None, repr=False
    )

    @staticmethod
    def _default_length_set_ranges(count, min_value, max_value):
        """
        Get default length set ranges.

        TODO: Currently not used.
        """
        arr = np.linspace(min_value, max_value, count + 1)
        starts = arr[0 : count + 1]
        ends = arr[1:]
        assert len(starts) == len(ends)
        as_gen = ((start, end) for start, end in zip(starts, ends))
        return tuple(as_gen)

    def __post_init__(self):
        """
        Copy GeoDataFrames instead of changing inputs.

        If the data is passed later to attribute, __setattr__ will also
        handle copying.

        :raises ValueError: If trace ``GeoDataFrame`` is empty after
            ``crop_to_target_areas``.
        """
        self.topology_determined = False
        if self.area_gdf.empty or self.area_gdf.geometry.iloc[0].is_empty:
            raise ValueError(
                "Passed area_gdf or geometry at index 0 is empty.\n"
                "Either pass a non-empty area_gdf or alternatively use\n"
                "fractopo.general.bounding_polygon to create an enveloping\n"
                "non-intersecting Polygon around your trace_gdf."
            )
        if self.circular_target_area:
            if not self.truncate_traces:
                raise ValueError(
                    "Traces must be truncated to the target area"
                    " to perform circular area trace weighting. "
                    "\n(To fix: pass truncate_traces=True.)"
                )
        # Copy geodataframes instead of using pointers
        # Traces
        self.trace_gdf = self.trace_gdf.copy()
        has_z_coordinates = check_for_z_coordinates(geodata=self.trace_gdf)
        if has_z_coordinates:
            log.warning(
                dedent(
                    """
                Traces inputted into Network contain Z-coordinates.
                They might not be correct in output geometries
                and might cause unexpected errors.
                """
                ).strip()
            )
            if self.remove_z_coordinates_from_inputs:
                log.warning(
                    dedent(
                        """
                    Z-coordinates will be removed from the input traces and
                    subsuquently from all outputs because Network input
                    remove_z_coordinates_from_geodata is True.
                    """
                    ).strip()
                )
                trace_gdf_without_z_coords = remove_z_coordinates_from_geodata(
                    geodata=self.trace_gdf
                )
                assert isinstance(trace_gdf_without_z_coords, gpd.GeoDataFrame)
                self.trace_gdf = trace_gdf_without_z_coords
        # Area
        self.area_gdf = self.area_gdf.copy()
        # Branches
        self.branch_gdf = self.branch_gdf.copy()
        # Branches
        self.node_gdf = self.node_gdf.copy()

        # The attribute trace_gdf by truncating it to the passed target
        # area
        if self.truncate_traces:
            self.trace_gdf = gpd.GeoDataFrame(
                crop_to_target_areas(
                    self.trace_gdf,
                    self.area_gdf,
                    keep_column_data=True,
                    allow_multilinestring_input=not self.determine_branches_nodes,
                )
            )
            self.trace_gdf.reset_index(inplace=True, drop=True)
            if self.trace_gdf.shape[0] == 0:
                raise ValueError("Empty trace GeoDataFrame after crop_to_target_areas.")

        self.trace_data = LineData(
            _line_gdf=self.trace_gdf,
            azimuth_set_ranges=self.azimuth_set_ranges,
            azimuth_set_names=self.azimuth_set_names,
            length_set_ranges=self.trace_length_set_ranges,
            length_set_names=self.trace_length_set_names,
            area_boundary_intersects=self.trace_intersects_target_area_boundary,
            using_branches=False,
        )

        empty_branches_and_nodes = self.branch_gdf.empty and self.node_gdf.empty
        if self.determine_branches_nodes and empty_branches_and_nodes:
            log.info(
                "Determining branches and nodes.",
                extra=dict(
                    empty_branches_and_nodes=empty_branches_and_nodes,
                    network_name=self.name,
                ),
            )
            self.assign_branches_nodes()
        elif not empty_branches_and_nodes:
            log.info(
                "Found branch_gdf and node_gdf in inputs. Using them.",
                extra=dict(
                    empty_branches_and_nodes=empty_branches_and_nodes,
                    network_name=self.name,
                ),
            )
            self.assign_branches_nodes(branches=self.branch_gdf, nodes=self.node_gdf)
            self.topology_determined = True

        log.info(
            "Created and initialized Network instance.",
            # Network has .name attribute which will overwrite log
            # attribute!!!
            extra={f"network_{key}": value for key, value in self.__dict__.items()},
        )

    def __hash__(self) -> int:
        """
        Implement Network hashing.

        Deprecated. Memory caching of ``GeoDataFrame``s is unstable.
        """

        def convert_gdf(
            gdf: Union[gpd.GeoDataFrame, gpd.GeoSeries, None, Polygon, MultiPolygon]
        ) -> Optional[str]:
            """
            Convert GeoDataFrame or geometry to (json) str.

            """
            if gdf is None:
                return None
            if isinstance(gdf, (gpd.GeoSeries, gpd.GeoDataFrame)):
                as_json = gdf.geometry.to_json(sort_keys=True)
                assert isinstance(as_json, str)
                return as_json
            as_wkt = gdf.wkt
            assert isinstance(as_wkt, str)
            return as_wkt

        traces_geojson = convert_gdf(self.trace_gdf)
        area_geojson = convert_gdf(self.area_gdf)

        hash_args = (
            traces_geojson,
            area_geojson,
            self.name,
            self.determine_branches_nodes,
            self.snap_threshold,
            self.truncate_traces,
            self.circular_target_area,
            self.azimuth_set_names,
            self.azimuth_set_ranges,
            self.trace_length_set_names,
            self.trace_length_set_ranges,
            self.branch_length_set_names,
            self.branch_length_set_ranges,
            convert_gdf(self.branch_gdf),
            convert_gdf(self.node_gdf),
            convert_gdf(self.censoring_area),
        )

        return hash(hash_args)

    def reset_length_data(self):
        """
        Reset LineData attributes.

        WARNING: Mostly untested.
        """
        log.warning(
            "Method Network.reset_length_data is mostly untested. Use with"
            "caution. It is safer to recreate the Network instance.",
            extra=dict(network_name=self.name),
        )
        # TODO: Refactor to use e.g. self.setup_trace_data as it is setup twice.
        # It is used here and in __post_init__
        self.trace_data = LineData(
            _line_gdf=self.trace_gdf,
            azimuth_set_ranges=self.azimuth_set_ranges,
            azimuth_set_names=self.azimuth_set_names,
            length_set_ranges=self.trace_length_set_ranges,
            length_set_names=self.trace_length_set_names,
            area_boundary_intersects=self.trace_intersects_target_area_boundary,
            using_branches=False,
        )
        if not self.branch_gdf.empty:
            self.branch_data = LineData(
                _line_gdf=self.branch_gdf,
                azimuth_set_ranges=self.azimuth_set_ranges,
                azimuth_set_names=self.azimuth_set_names,
                length_set_ranges=self.branch_length_set_ranges,
                length_set_names=self.branch_length_set_names,
                area_boundary_intersects=self.branch_intersects_target_area_boundary,
                using_branches=True,
            )
            self._azimuth_set_relationships = None

    @property
    def plain_name(self):
        """
        Get filename friendly name for Network based on ``name`` attribute.
        """
        return sanitize_name(self.name)

    @property
    def trace_series(self) -> gpd.GeoSeries:
        """
        Get trace geometries as GeoSeries.
        """
        return self.trace_data.geometry

    @property
    @requires_topology
    def node_series(self) -> gpd.GeoSeries:
        """
        Get node geometries as GeoSeries.
        """
        return self.node_gdf.geometry

    @property
    @requires_topology
    def branch_series(self) -> gpd.GeoSeries:
        """
        Get branch geometries as GeoSeries.
        """
        return self.branch_data.geometry

    @property
    def trace_azimuth_array(self) -> np.ndarray:
        """
        Get trace azimuths as array.
        """
        return self.trace_data.azimuth_array

    @property
    @requires_topology
    def branch_azimuth_array(self) -> np.ndarray:
        """
        Get branch azimuths as array.
        """
        return self.branch_data.azimuth_array

    @property
    def trace_length_array(self) -> np.ndarray:
        """
        Get trace lengths as array.
        """
        return self.trace_data.length_array

    @property
    def trace_length_array_non_weighted(self) -> np.ndarray:
        """
        Get non-boundary-weighted trace lengths as array.
        """
        return self.trace_data.length_array_non_weighted

    @property
    @requires_topology
    def branch_length_array(self) -> np.ndarray:
        """
        Get branch lengths as array.
        """
        return self.branch_data.length_array

    @property
    @requires_topology
    def branch_length_array_non_weighted(self) -> np.ndarray:
        """
        Get non-boundary-weighted branch lengths as array.
        """
        return self.branch_data.length_array_non_weighted

    @property
    def trace_azimuth_set_array(self) -> np.ndarray:
        """
        Get azimuth set for each trace.
        """
        return self.trace_data.azimuth_set_array

    @property
    @requires_topology
    def branch_azimuth_set_array(self) -> np.ndarray:
        """
        Get azimuth set for each branch.
        """
        return self.branch_data.azimuth_set_array

    @property
    def trace_length_set_array(self) -> np.ndarray:
        """
        Get length set for each trace.
        """
        return self.trace_data.length_set_array

    @property
    @requires_topology
    def branch_length_set_array(self) -> np.ndarray:
        """
        Get length set for each branch.
        """
        return self.branch_data.length_set_array

    @property
    @requires_topology
    def node_types(self) -> np.ndarray:
        """
        Get node type of each node.
        """
        node_class_series = self.node_gdf[CLASS_COLUMN]
        assert isinstance(node_class_series, pd.Series)
        numpy_array = node_class_series.to_numpy()
        assert isinstance(numpy_array, np.ndarray)
        return numpy_array

    @property
    def node_counts(self) -> Dict[str, int]:
        """
        Get node counts.
        """
        return convert_counts(
            determine_node_type_counts(
                self.node_types, branches_defined=self.topology_determined
            )
        )

    @property
    @requires_topology
    def branch_types(self) -> np.ndarray:
        """
        Get branch type of each branch.
        """
        branch_connection_series = self.branch_gdf[CONNECTION_COLUMN]
        assert isinstance(branch_connection_series, pd.Series)
        as_array = branch_connection_series.to_numpy()
        assert isinstance(as_array, np.ndarray)
        return as_array

    @property
    @requires_topology
    def branch_counts(self) -> Dict[str, int]:
        """
        Get branch counts.
        """
        return convert_counts(
            determine_branch_type_counts(
                self.branch_types, branches_defined=self.topology_determined
            )
        )

    @property
    def total_area(self) -> float:
        """
        Get total area.
        """
        area_sum = numpy_to_python_type(self.area_gdf.geometry.area.sum())
        assert isinstance(area_sum, float)
        assert area_sum >= 0.0
        return area_sum

    @property
    def parameters(self) -> Dict[str, float]:
        """
        Get numerical geometric and topological parameters.
        """
        # Cannot do simple cached_property because None might have been
        # returned previously.
        if self._parameters is None:
            branch_length_array = (
                self.branch_length_array_non_weighted
                if self.topology_determined
                else None
            )
            node_counts = self.node_counts if self.topology_determined else None
            self._parameters = determine_topology_parameters(
                trace_length_array=self.trace_length_array_non_weighted,
                area=self.total_area,
                branches_defined=self.topology_determined,
                correct_mauldon=self.circular_target_area,
                # Require determined topology (branches & nodes)
                node_counts=node_counts,
                branch_length_array=branch_length_array,
            )
        return self._parameters

    @property
    @requires_topology
    def anisotropy(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Determine anisotropy of connectivity.
        """
        if self._anisotropy is None:
            self._anisotropy = determine_anisotropy_sum(
                azimuth_array=self.branch_azimuth_array,
                length_array=self.branch_length_array,
                branch_types=self.branch_types,
            )
        return self._anisotropy

    @property
    @requires_topology
    def azimuth_set_relationships(self) -> pd.DataFrame:
        """
        Determine azimuth set relationships.
        """
        if self._azimuth_set_relationships is None:
            self._azimuth_set_relationships = determine_crosscut_abutting_relationships(
                trace_series=self.trace_series,
                node_series=self.node_series,  # type: ignore
                node_types=self.node_types,
                set_array=self.trace_azimuth_set_array,
                set_names=self.azimuth_set_names,
                buffer_value=0.001,
                label=self.name,
            )
        return self._azimuth_set_relationships

    @property
    @requires_topology
    def length_set_relationships(self) -> pd.DataFrame:
        """
        Determine length set relationships.
        """
        if self._trace_length_set_relationships is None:
            self._trace_length_set_relationships = (
                determine_crosscut_abutting_relationships(
                    trace_series=self.trace_series,
                    node_series=self.node_series,  # type: ignore
                    node_types=self.node_types,
                    set_array=self.trace_data.length_set_array,
                    set_names=self.trace_data.length_set_names,  # type: ignore
                    buffer_value=0.001,
                    label=self.name,
                )
            )
        return self._trace_length_set_relationships

    @property
    def trace_azimuth_set_counts(self) -> Dict[str, int]:
        """
        Get trace azimuth set counts.
        """
        return self.trace_data.azimuth_set_counts

    @property
    def trace_length_set_counts(self) -> Dict[str, int]:
        """
        Get trace length set counts.
        """
        return self.trace_data.length_set_counts

    @property
    def branch_azimuth_set_counts(self) -> Dict[str, int]:
        """
        Get branch azimuth set counts.
        """
        return self.branch_data.azimuth_set_counts

    @property
    def branch_length_set_counts(self) -> Dict[str, int]:
        """
        Get branch length set counts.
        """
        return self.branch_data.length_set_counts

    @property
    def trace_lengths_powerlaw_fit_description(self) -> Dict[str, float]:
        """
        Short numerical description dict of trace length powerlaw fit.
        """
        return self.trace_data.describe_fit(label="trace")

    @property
    def branch_lengths_powerlaw_fit_description(self) -> Dict[str, float]:
        """
        Short numerical description dict of branch length powerlaw fit.
        """
        return self.branch_data.describe_fit(label="branch")

    @property
    def target_areas(self) -> List[Union[Polygon, MultiPolygon]]:
        """
        Get all target areas from area_gdf.
        """
        target_areas = []
        for target_area in self.area_gdf.geometry.values:
            if not isinstance(target_area, (Polygon, MultiPolygon)):
                raise TypeError("Expected (Multi)Polygon geometries in area_gdf.")
            if target_area.is_empty:
                raise ValueError("Expected non-empty geometries for target areas.")
            target_areas.append(target_area)
        return target_areas

    @property
    def trace_intersects_target_area_boundary(self) -> np.ndarray:
        """
        Check traces for intersection with target area boundaries.

        Results are in integers:

        -  0 == No intersections
        -  1 == One intersection
        -  2 == Two intersections

        Does not discriminate between which target area (if multiple) the trace
        intersects. Intersection detection based on snap_threshold.
        """
        if self._trace_intersects_target_area_boundary is None:
            if self.circular_target_area:
                (
                    intersecting_lines,
                    cuts_through_lines,
                ) = determine_boundary_intersecting_lines(
                    line_gdf=self.trace_gdf,
                    area_gdf=self.area_gdf,
                    snap_threshold=self.snap_threshold,
                )
                self._trace_intersects_target_area_boundary = bool_arrays_sum(
                    intersecting_lines, cuts_through_lines
                )
            else:
                self._trace_intersects_target_area_boundary = np.array(
                    [0] * self.trace_gdf.shape[0]
                )
        return self._trace_intersects_target_area_boundary

    @property
    @requires_topology
    def branch_intersects_target_area_boundary(self) -> np.ndarray:
        """
        Get array of E-component count.
        """
        if self._branch_intersects_target_area_boundary is None:
            if self.circular_target_area:
                intersecting_lines = branches_intersect_boundary(self.branch_types)
                cuts_through_lines = np.array(
                    [branch_type == EE_branch for branch_type in self.branch_types]
                )
                self._branch_intersects_target_area_boundary = bool_arrays_sum(
                    intersecting_lines, cuts_through_lines
                )
            else:
                self._branch_intersects_target_area_boundary = np.array(
                    [0] * len(self.branch_types)
                )
        return self._branch_intersects_target_area_boundary

    @requires_topology
    def numerical_network_description(
        self,
        trace_lengths_cut_off: Optional[float] = None,
        branch_lengths_cut_off: Optional[float] = None,
    ) -> Dict[str, Union[Number, str]]:
        """
        Collect numerical network attributes and return them as a dictionary.
        """
        parameters = self.parameters
        branch_counts = self.branch_counts
        node_counts = self.node_counts
        trace_boundary_intersect_count = self.trace_data.boundary_intersect_count_desc(
            label="Trace"
        )
        branch_boundary_intersect_count = (
            self.branch_data.boundary_intersect_count_desc(label="Branch")
        )
        radius = {
            RADIUS: (
                calc_circle_radius(parameters[Param.AREA.value.name])
                if self.circular_target_area
                else np.nan
            )
        }
        censoring_value = self.estimate_censoring()
        censoring_and_relative = {
            CENSORING: censoring_value,
            RELATIVE_CENSORING: censoring_value / parameters[Param.AREA.value.name],
        }
        description = {
            **trace_boundary_intersect_count,
            **branch_boundary_intersect_count,
            **branch_counts,
            **node_counts,
            **parameters,
            **radius,
            **self.trace_data.describe_fit(
                label="trace", cut_off=trace_lengths_cut_off
            ),
            **self.branch_data.describe_fit(
                label="branch", cut_off=branch_lengths_cut_off
            ),
            NAME: self.name,
            REPRESENTATIVE_POINT: MultiPoint(self.representative_points()).centroid.wkt,
            **censoring_and_relative,
        }
        assert isinstance(description, dict)

        return description

    def representative_points(self) -> List[Point]:
        """
        Get representative point(s) of target area(s).
        """
        point_list = self.area_gdf.representative_point().to_list()
        assert isinstance(point_list, list)
        assert all(isinstance(point, Point) for point in point_list)
        return point_list

    def trace_lengths_powerlaw_fit(
        self, cut_off: Optional[float] = None
    ) -> Optional[powerlaw.Fit]:
        """
        Determine powerlaw fit for trace lengths.
        """
        return (
            self.trace_data.automatic_fit
            if cut_off is None
            else self.trace_data.determine_manual_fit(cut_off=cut_off)
        )

    @requires_topology
    def branch_lengths_powerlaw_fit(
        self, cut_off: Optional[float] = None
    ) -> Optional[powerlaw.Fit]:
        """
        Determine powerlaw fit for branch lengths.
        """
        return (
            self.branch_data.automatic_fit
            if cut_off is None
            else self.branch_data.determine_manual_fit(cut_off=cut_off)
        )

    def assign_branches_nodes(
        self,
        branches: Optional[gpd.GeoDataFrame] = None,
        nodes: Optional[gpd.GeoDataFrame] = None,
    ):
        """
        Determine and assign branches and nodes as attributes.
        """
        if branches is None or nodes is None:
            branches, nodes = branches_and_nodes(
                traces=self.trace_gdf,
                areas=self.area_gdf,
                snap_threshold=self.snap_threshold,
                already_clipped=self.truncate_traces,
                # unary_size_threshold=self.unary_size_threshold,
            )
        else:
            branches = branches.copy()
            nodes = nodes.copy()
        # if self.trace_gdf.crs is not None:
        #     branches.set_crs(self.trace_gdf.crs, inplace=True)
        #     nodes.set_crs(self.trace_gdf.crs, inplace=True)
        self.branch_gdf = branches
        self.node_gdf = nodes
        self.topology_determined = True
        self.branch_data = LineData(
            _line_gdf=self.branch_gdf,
            azimuth_set_ranges=self.azimuth_set_ranges,
            azimuth_set_names=self.azimuth_set_names,
            length_set_ranges=self.branch_length_set_ranges,
            length_set_names=self.branch_length_set_names,
            area_boundary_intersects=self.branch_intersects_target_area_boundary,
            using_branches=True,
        )

    def plot_trace_lengths(
        self,
        label: Optional[str] = None,
        fit: Optional[powerlaw.Fit] = None,
        use_probability_density_function: bool = False,
        plain: bool = False,
    ) -> Tuple[powerlaw.Fit, Figure, Axes]:  # type: ignore
        """
        Plot trace length distribution with `powerlaw` fits.
        """
        label = self.name if label is None else label
        return self.trace_data.plot_lengths(
            label=label,
            fit=fit,
            use_probability_density_function=use_probability_density_function,
            plain=plain,
        )

    @requires_topology
    def plot_branch_lengths(
        self,
        label: Optional[str] = None,
        fit: Optional[powerlaw.Fit] = None,
        use_probability_density_function: bool = False,
        plain: bool = False,
    ) -> Tuple[Optional[powerlaw.Fit], Figure, Axes]:  # type: ignore
        """
        Plot branch length distribution with `powerlaw` fits.
        """
        if label is None:
            label = self.name
        return self.branch_data.plot_lengths(
            label=label,
            fit=fit,
            use_probability_density_function=use_probability_density_function,
            plain=plain,
        )

    def plot_trace_azimuth(
        self,
        label: Optional[str] = None,
        append_azimuth_set_text: bool = False,
        add_abundance_order: bool = False,
        visualize_sets: bool = False,
        bar_color: str = "darkgrey",
        plain: bool = False,
    ) -> Tuple[AzimuthBins, Figure, PolarAxes]:
        """
        Plot trace azimuth rose plot.
        """
        if label is None:
            label = self.name
        return self.trace_data.plot_azimuth(
            label=label,
            append_azimuth_set_text=append_azimuth_set_text,
            add_abundance_order=add_abundance_order,
            visualize_sets=visualize_sets,
            bar_color=bar_color,
            plain=plain,
        )

    @requires_topology
    def plot_branch_azimuth(
        self,
        label: Optional[str] = None,
        append_azimuth_set_text: bool = False,
        add_abundance_order: bool = False,
        visualize_sets: bool = False,
        bar_color: str = "darkgrey",
        plain: bool = False,
    ) -> Tuple[AzimuthBins, Figure, PolarAxes]:
        """
        Plot branch azimuth rose plot.
        """
        if label is None:
            label = self.name
        return self.branch_data.plot_azimuth(
            label=label,
            append_azimuth_set_text=append_azimuth_set_text,
            add_abundance_order=add_abundance_order,
            visualize_sets=visualize_sets,
            bar_color=bar_color,
            plain=plain,
        )

    @requires_topology
    def plot_xyi(
        self, label: Optional[str] = None
    ) -> Tuple[Figure, Axes, TernaryAxesSubplot]:
        """
        Plot ternary plot of node types.
        """
        if label is None:
            label = self.name
        return plot_ternary_plot(
            counts_list=[self.node_counts], labels=[label], is_nodes=True
        )

    @requires_topology
    def plot_branch(
        self, label: Optional[str] = None
    ) -> Tuple[Figure, Axes, TernaryAxesSubplot]:
        """
        Plot ternary plot of branch types.
        """
        if label is None:
            label = self.name
        return plot_ternary_plot(
            counts_list=[self.branch_counts], labels=[label], is_nodes=False
        )

    def plot_parameters(
        self, label: Optional[str] = None, color: Optional[str] = None
    ) -> Optional[Tuple[Figure, Axes]]:
        """
        Plot geometric and topological parameters.
        """
        if label is None:
            label = self.name
        if color is None:
            color = "black"
        assert self.parameters is not None
        figs, axes = plot_parameters_plot(
            topology_parameters_list=[self.parameters],  # type: ignore
            labels=[label],
            colors=[color],
        )
        return figs[0], axes[0]

    @requires_topology
    def plot_anisotropy(
        self, label: Optional[str] = None, color: Optional[str] = None
    ) -> Optional[Tuple[Figure, Axes]]:
        """
        Plot anisotropy of connectivity plot.
        """
        if label is None:
            label = self.name
        if color is None:
            color = "black"
        anisotropy_sum = self.anisotropy[0]
        sample_intervals = self.anisotropy[1]
        fig, ax = plot_anisotropy_plot(
            anisotropy_sum=anisotropy_sum,
            sample_intervals=sample_intervals,
            # label=label,  # type: ignore
            # color=color,
        )
        return fig, ax

    @requires_topology
    def plot_azimuth_crosscut_abutting_relationships(
        self,
    ) -> Tuple[List[Figure], List[np.ndarray]]:
        """
        Plot azimuth set crosscutting and abutting relationships.
        """
        return plot_crosscut_abutting_relationships_plot(
            relations_df=self.azimuth_set_relationships,  # type: ignore
            set_array=self.trace_azimuth_set_array,
            set_names=self.azimuth_set_names,
        )

    @requires_topology
    def plot_trace_length_crosscut_abutting_relationships(
        self,
    ) -> Tuple[List[Figure], List[np.ndarray]]:
        """
        Plot length set crosscutting and abutting relationships.
        """
        return plot_crosscut_abutting_relationships_plot(
            relations_df=self.length_set_relationships,  # type: ignore
            set_array=self.trace_data.length_set_array,
            set_names=self.trace_data.length_set_names,  # type: ignore
        )

    def plot_trace_azimuth_set_count(
        self, label: Optional[str] = None
    ) -> Tuple[Figure, Axes]:
        """
        Plot trace azimuth set counts.
        """
        if label is None:
            label = self.name
        return self.trace_data.plot_azimuth_set_count(label=label)

    @requires_topology
    def plot_branch_azimuth_set_count(
        self, label: Optional[str] = None
    ) -> Tuple[Figure, Axes]:
        """
        Plot branch azimuth set counts.
        """
        if label is None:
            label = self.name
        return self.branch_data.plot_azimuth_set_count(label=label)

    def plot_trace_length_set_count(
        self, label: Optional[str] = None
    ) -> Tuple[Figure, Axes]:
        """
        Plot trace length set counts.
        """
        if label is None:
            label = self.name
        return self.trace_data.plot_length_set_count(label=label)

    def plot_branch_length_set_count(
        self, label: Optional[str] = None
    ) -> Tuple[Figure, Axes]:
        """
        Plot branch length set counts.
        """
        if label is None:
            label = self.name
        return self.branch_data.plot_length_set_count(label=label)

    def plot_trace_azimuth_set_lengths(
        self,
    ) -> Tuple[List[Optional[powerlaw.Fit]], List[Figure], List[Axes]]:
        """
        Plot trace azimuth set lengths with fits.
        """
        return self.trace_data.plot_azimuth_set_lengths()

    @requires_topology
    def plot_branch_azimuth_set_lengths(
        self,
    ) -> Tuple[List[Optional[powerlaw.Fit]], List[Figure], List[Axes]]:
        """
        Plot branch azimuth set lengths with fits.
        """
        return self.branch_data.plot_azimuth_set_lengths()

    def contour_grid(
        self,
        cell_width: Optional[float] = None,
        bounds_divider: float = 20.0,
        precursor_grid: Optional[gpd.GeoDataFrame] = None,
        resolve_branches_nodes: bool = False,
    ):
        """
        Sample the network with a contour grid.

        If ``cell_width`` is passed it is used as the cell width. Otherwise
        a cell width is calculated using the network branch bounds using
        the passed ``bounds_divider`` or its default value.

        If ``precursor_grid`` is passed it is used as the grid in which
        each Polygon cell is filled with calculated network parameter values.
        """
        line_gdf = self.branch_gdf if self.determine_branches_nodes else self.trace_gdf
        if cell_width is None:
            # Use trace bounds to calculate a width
            min_x, min_y, max_x, max_y = total_bounds(line_gdf)
            x_diff = max_x - min_x
            y_diff = max_y - min_y
            if x_diff < y_diff:
                cell_width = y_diff / bounds_divider
            else:
                cell_width = x_diff / bounds_divider

        assert isinstance(cell_width, (float, int))

        sampled_grid = run_grid_sampling(
            traces=self.trace_gdf,
            branches=self.branch_gdf,
            nodes=self.node_gdf,
            cell_width=cell_width,
            snap_threshold=self.snap_threshold,
            precursor_grid=precursor_grid,
            resolve_branches_and_nodes=resolve_branches_nodes,
        )
        return sampled_grid

    def plot_contour(
        self, parameter: str, sampled_grid: gpd.GeoDataFrame
    ) -> Tuple[Figure, Axes]:
        """
        Plot contour plot of a geometric or topological parameter.

        Creating the contour grid is expensive so the ``sampled_grid`` must
        be first created with ``Network.contour_grid`` method and then passed
        to this one for plotting.
        """
        assert all(isinstance(val, Polygon) for val in sampled_grid.geometry.values)
        fig, ax = plt.subplots(1, 1, figsize=(12, 12))
        assert isinstance(fig, Figure)
        assert isinstance(ax, Axes)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        sampled_grid.plot(
            column=parameter,
            legend=True,
            cax=cax,
            ax=ax,
            legend_kwds={"label": parameter},
        )
        ax.set_title(f"{ self.name } - {parameter}")
        return fig, ax

    def estimate_censoring(
        self,
    ) -> float:
        """
        Estimate the amount of censoring as area float value.

        Censoring is caused by e.g. vegetation.

        Returns np.nan if no ``censoring_area`` is passed by the user into
        ``Network`` creation or if the passed GeoDataFrame is empty.
        """
        # Either censoring_area is passed directly or its passed in Network
        # creation. Otherwise raise ValueError.
        if self.censoring_area is None:
            raise ValueError(
                "Expected censoring_area as an argument or initialized"
                f" as Network (name:{self.name}) attribute."
            )
        if not isinstance(self.censoring_area, gpd.GeoDataFrame):
            raise TypeError(
                "Expected censoring_area to be of type gpd.GeoDataFrame."
                f" Got type={type(self.censoring_area)}"
            )
        if self.censoring_area.empty:
            return np.nan

        # Determine bounds of Network.area_gdf
        network_area_bounds = total_bounds(self.area_gdf)

        # Use spatial index to filter censoring polygons that are not near the
        # network
        sindex = pygeos_spatial_index(self.censoring_area)
        index_intersection = spatial_index_intersection(
            spatial_index=sindex, coordinates=network_area_bounds
        )
        candidate_idxs = list(
            index_intersection if index_intersection is not None else []
        )
        if len(candidate_idxs) == 0:
            return 0.0
        candidates = self.censoring_area.iloc[candidate_idxs]

        # Clip the censoring areas with the network area and calculate the area
        # sum of leftover area.
        clipped = gpd.clip(candidates, self.area_gdf)
        if not isinstance(clipped, (gpd.GeoDataFrame, gpd.GeoSeries)):
            vals = type(clipped), clipped
            raise TypeError(
                f"Expected that clipped is of geopandas data type. Got: {vals}."
            )
        censoring_value = clipped.area.sum()
        unpacked_value = numpy_to_python_type(censoring_value)
        assert isinstance(unpacked_value, float)
        assert unpacked_value >= 0.0
        return unpacked_value

    @requires_topology
    def write_branches_and_nodes(
        self,
        output_dir_path: Path,
        branches_name: Optional[str] = None,
        nodes_name: Optional[str] = None,
    ):
        """
        Write branches and nodes to disk.

        Enables reuse of the same data in analysis of the same data to skip
        topology determination which is computationally expensive.

        Writes only with the GeoJSON driver as there are differences between
        different spatial filetypes. Only GeoJSON is supported to avoid unexpected
        errors.
        """
        for topo_name, topo_type, gdf in zip(
            (branches_name, nodes_name),
            ("branches", "nodes"),
            (self.branch_gdf, self.node_gdf),
        ):
            topo_path = output_dir_path / (
                f"{self.name}_{topo_type}.geojson" if topo_name is None else topo_name
            )
            write_geodata(gdf=gdf, path=topo_path, allow_list_column_transform=False)
            log.info(
                "Wrote topological data to disk.",
                extra=dict(
                    network_name=self.name,
                    topo_path=topo_path,
                    topo_type=topo_type,
                ),
            )

    def _length_distribution(
        self, using_branches: bool, azimuth_set: Optional[str]
    ) -> LengthDistribution:
        """
        Create structured LengthDistribution instance.
        """
        # set_array = (
        #     self.branch_azimuth_set_array
        #     if using_branches
        #     else self.trace_azimuth_set_array
        # )
        if azimuth_set is None:
            lengths = (
                self.branch_length_array if using_branches else self.trace_length_array
            )
            name = self.name
        else:
            lengths = (
                self.branch_data.azimuth_set_length_arrays
                if using_branches
                else self.trace_data.azimuth_set_length_arrays
            )[azimuth_set]
            # lengths = length_array[set_array == azimuth_set]
            name = f"{self.name} {azimuth_set}"
        return LengthDistribution(
            lengths=lengths,
            area_value=self.total_area,
            name=name,
            using_branches=using_branches,
        )

    def trace_length_distribution(
        self, azimuth_set: Optional[str]
    ) -> LengthDistribution:
        """
        Create structured LengthDistribution instance of trace length data.
        """
        return self._length_distribution(using_branches=False, azimuth_set=azimuth_set)

    @requires_topology
    def branch_length_distribution(
        self, azimuth_set: Optional[str]
    ) -> LengthDistribution:
        """
        Create structured LengthDistribution instance of branch length data.
        """
        return self._length_distribution(using_branches=True, azimuth_set=azimuth_set)

    @requires_topology
    def export_network_analysis(
        self, output_path: Path, include_contour_grid: bool = True
    ):
        """
        Export pre-selected ``Network`` analysis results to a directory.

        The chosen analyses are opionated but should contain at least the basic
        results of fracture network analysis.

        ``output_path`` should correspond to a path to an existing or
        directory or direct path to a non-existing directory where one will
        be created.
        """
        if output_path.is_dir():
            export_path = output_path / self.plain_name
        elif not output_path.exists():
            export_path = output_path
        else:
            raise ValueError(
                "Expected output_path to correspond to a path to a "
                "directory or a path to where an export directory will be created."
            )
        export_path.mkdir(exist_ok=False)

        # Plot map of traces with area
        fig, ax = plt.subplots(figsize=(9, 9))
        self.trace_gdf.plot(ax=ax, color="blue")
        self.area_gdf.boundary.plot(ax=ax, color="red")
        focus_plot_to_bounds(ax, self.area_gdf.total_bounds)
        save_fig(fig=fig, results_dir=export_path, name="trace_map")

        # Plot map of branches
        fig, ax = plt.subplots(figsize=(9, 9))
        self.branch_gdf.plot(
            colors=[assign_branch_and_node_colors(bt) for bt in self.branch_types],
            ax=ax,
        )
        self.area_gdf.boundary.plot(ax=ax, color="red")
        save_fig(fig=fig, results_dir=export_path, name="branches_map")

        # Plot map of nodes
        fig, ax = plt.subplots(figsize=(9, 9))
        # Traces
        self.trace_gdf.plot(ax=ax, linewidth=0.5)
        # Nodes
        self.node_gdf.plot(
            c=[assign_branch_and_node_colors(bt) for bt in self.node_types],
            ax=ax,
            markersize=10,
        )
        self.area_gdf.boundary.plot(ax=ax, color="red")
        focus_plot_to_bounds(ax, self.area_gdf.total_bounds)
        save_fig(fig=fig, results_dir=export_path, name="nodes_map")

        # Plot azimuth rose plot of fracture traces
        _, fig, _ = self.plot_trace_azimuth()
        save_fig(
            fig=fig, results_dir=export_path, name="trace_length_weighted_rose_plot"
        )

        # Plot length distribution fits of fracture traces
        _, fig, _ = self.plot_trace_lengths()
        save_fig(
            fig=fig, results_dir=export_path, name="trace_length_distribution_fits"
        )

        # Plot length distribution fits of fracture branches
        _, fig, _ = self.plot_branch_lengths()
        save_fig(
            fig=fig, results_dir=export_path, name="branch_length_distribution_fits"
        )

        # Plot crosscutting and abutting relationships between azimuth sets
        figs, _ = self.plot_azimuth_crosscut_abutting_relationships()
        for i, fig in enumerate(figs):
            save_fig(
                fig=fig, results_dir=export_path, name=f"azimuth_set_relationships_{i}"
            )

        # Plot ternary XYI-node proportion plot
        fig, _, _ = self.plot_xyi()
        save_fig(fig=fig, results_dir=export_path, name="xyi_ternary_plot")

        # Plot ternary branch (C-C, C-I, I-I) proportion plot
        fig, _, _ = self.plot_branch()
        save_fig(fig=fig, results_dir=export_path, name="branch_ternary_plot")

        # Save traces, branches and nodes.
        write_geodataframe(
            self.trace_gdf, f"{self.plain_name}_traces", results_dir=export_path
        )
        write_geodataframe(
            self.branch_gdf, f"{self.plain_name}_branches", results_dir=export_path
        )
        write_geodataframe(
            self.node_gdf, f"{self.plain_name}_nodes", results_dir=export_path
        )

        # Create point with general parameters of the Network
        params_point = gpd.GeoDataFrame(
            [
                {
                    **self.parameters,
                    "geometry": MultiPoint(
                        self.area_gdf.geometry.representative_point()
                    ).representative_point(),
                }
            ],
            crs=self.area_gdf.crs,
        )
        write_geodataframe(
            params_point, f"{self.plain_name}_parameters", results_dir=export_path
        )

        # Create contour grid
        if include_contour_grid:
            sampled_grid = self.contour_grid()
            write_geodataframe(
                sampled_grid, f"{self.plain_name}_contour_grid", results_dir=export_path
            )
            fig, _ = self.plot_contour(
                Param.FRACTURE_INTENSITY_P21.value.name, sampled_grid=sampled_grid
            )
            save_fig(fig=fig, results_dir=export_path, name="P21_contour")

            fig, _ = self.plot_contour(
                Param.CONNECTIONS_PER_BRANCH.value.name, sampled_grid=sampled_grid
            )
            save_fig(
                fig=fig, results_dir=export_path, name="connections_per_branch_contour"
            )



"""
Determine traces that could be integrated together.

determine_proximal_traces takes an input of GeoSeries or GeoDataFrame
of LineString geometries and returns a GeoDataFrame with a new
column `Merge` which has values of True or False depending on if
nearby proximal traces were found.
"""
from typing import List, Union

import geopandas as gpd
from shapely.geometry import LineString

from fractopo.general import (
    determine_azimuth,
    determine_regression_azimuth,
    geom_bounds,
    is_azimuth_close,
    pygeos_spatial_index,
    safe_buffer,
    spatial_index_intersection,
)

MERGE_COLUMN = "Merge"


def is_within_buffer_distance(
    trace: LineString, other: LineString, buffer_value: float
):
    """
    Determine if `trace` and `other` are within buffer distance.

    Threshold distance is buffer_value (both are buffered with half of
    buffer_value) of each other.

    E.g.

    >>> trace = LineString([(0, 0), (0, 3)])
    >>> other = LineString([(0, 0), (0, 3)])
    >>> is_within_buffer_distance(trace, other, 0.1)
    True

    >>> line = LineString([(0, 0), (0, 3)])
    >>> other = LineString([(3, 0), (3, 3)])
    >>> is_within_buffer_distance(trace, other, 2)
    False
    >>> is_within_buffer_distance(trace, other, 4)
    True

    """
    return trace.buffer(buffer_value / 2).intersects(other.buffer(buffer_value / 2))


def is_similar_azimuth(trace: LineString, other: LineString, tolerance: float):
    """
    Determine if azimuths of `trace` and `other` are close.

    Checks both the start -- end -azimuth and regression-based azimuth.

    E.g.

    >>> trace = LineString([(0, 0), (0, 3)])
    >>> other = LineString([(0, 0), (0, 4)])
    >>> is_similar_azimuth(trace, other, 1)
    True

    >>> trace = LineString([(0, 0), (1, 1)])
    >>> other = LineString([(0, 0), (0, 4)])
    >>> is_similar_azimuth(trace, other, 40)
    False
    >>> is_similar_azimuth(trace, other, 50)
    True
    >>> is_similar_azimuth(trace, other, 45)
    False

    """
    return is_azimuth_close(
        determine_azimuth(trace, halved=True),
        determine_azimuth(other, halved=True),
        tolerance=tolerance,
        halved=True,
    ) or is_azimuth_close(
        determine_regression_azimuth(trace),
        determine_regression_azimuth(other),
        tolerance=tolerance,
        halved=True,
    )


def determine_proximal_traces(
    traces: Union[gpd.GeoSeries, gpd.GeoDataFrame],
    buffer_value: float,
    azimuth_tolerance: float,
) -> gpd.GeoDataFrame:
    """
    Determine proximal traces.

    Takes an input of GeoSeries or GeoDataFrame
    of LineString geometries and returns a GeoDataFrame with a new
    column `Merge` which has values of True or False depending on if
    nearby proximal traces were found.

    E.g.

    >>> lines = [
    ...     LineString([(0, 0), (0, 3)]),
    ...     LineString([(1, 0), (1, 3)]),
    ...     LineString([(5, 0), (5, 3)]),
    ...     LineString([(0, 0), (-3, -3)]),
    ... ]
    >>> traces = gpd.GeoDataFrame({"geometry": lines})
    >>> buffer_value = 1.1
    >>> azimuth_tolerance = 10
    >>> determine_proximal_traces(traces, buffer_value, azimuth_tolerance)
                                              geometry  Merge
    0    LINESTRING (0.00000 0.00000, 0.00000 3.00000)   True
    1    LINESTRING (1.00000 0.00000, 1.00000 3.00000)   True
    2    LINESTRING (5.00000 0.00000, 5.00000 3.00000)  False
    3  LINESTRING (0.00000 0.00000, -3.00000 -3.00000)  False

    """
    assert isinstance(traces, (gpd.GeoSeries, gpd.GeoDataFrame))
    if isinstance(traces, gpd.GeoSeries):
        traces_as_gdf: gpd.GeoDataFrame = gpd.GeoDataFrame(geometry=traces)
    else:
        traces_as_gdf = traces
    traces_as_gdf.reset_index(inplace=True, drop=True)
    spatial_index = pygeos_spatial_index(traces_as_gdf)
    trace: LineString
    proximal_traces: List[int] = []
    for idx, trace in enumerate(traces_as_gdf.geometry.values):
        candidate_idxs = spatial_index_intersection(
            spatial_index, geom_bounds(safe_buffer(trace, buffer_value * 5))
        )
        candidate_idxs.remove(idx)
        candidate_traces: Union[gpd.GeoSeries, gpd.GeoDataFrame] = traces_as_gdf.iloc[
            candidate_idxs
        ]
        candidate_traces = candidate_traces.loc[  # type: ignore
            [
                is_within_buffer_distance(trace, other, buffer_value)
                and is_similar_azimuth(trace, other, tolerance=azimuth_tolerance)
                for other in candidate_traces.geometry.values
            ]
        ]
        if len(candidate_traces) > 0:
            proximal_traces.extend(
                [
                    i
                    for i in list(candidate_traces.index) + [idx]  # type: ignore
                    if i not in proximal_traces
                ]
            )
    traces_as_gdf[MERGE_COLUMN] = [i in proximal_traces for i in traces_as_gdf.index]
    return traces_as_gdf



"""
Functions for plotting rose plots.
"""

import math
from dataclasses import dataclass
from textwrap import fill
from typing import Optional, Tuple

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.projections import PolarAxes

from fractopo.general import Number, SetRangeTuple


@dataclass
class AzimuthBins:

    """
    Dataclass for azimuth rose plot bin data.
    """

    bin_width: float
    bin_locs: np.ndarray
    bin_heights: np.ndarray


def _calc_ideal_bin_width(n: Number, axial=True) -> float:
    """
    Calculate ideal bin width. axial or vector data.

    Reference:
        Sanderson, D.J., Peacock, D.C.P., 2020.
        Making rose diagrams fit-for-purpose. Earth-Science Reviews.
        doi:10.1016/j.earscirev.2019.103055

    E.g.

    >>> _calc_ideal_bin_width(30)
    28.964681538168897

    >>> _calc_ideal_bin_width(90)
    20.08298850246509

    :param n: Sample size
    :type n: int
    :param axial: Whether data is axial or vector
    :type axial: bool
    :return: Bin width in degrees
    :raises: ValueError
    """
    if n <= 0:
        raise ValueError("Sample size cannot be 0 or lower")
    if axial:
        degree_range = 180
    else:
        degree_range = 360
    result = degree_range / (2 * n ** (1 / 3))
    assert isinstance(result, float)
    return result


def _calc_bins(ideal_bin_width: float, axial: bool) -> Tuple[np.ndarray, float]:
    """
    Calculate bin edges and real bin width from ideal bin width.

    E.g.

    >>> _calc_bins(25.554235, True)
    (array([  0. ,  22.5,  45. ,  67.5,  90. , 112.5, 135. , 157.5, 180. ]), 22.5)
    """
    max_angle = 180 if axial else 360
    div = max_angle / ideal_bin_width
    rounded_div = math.ceil(div)
    bin_width = max_angle / rounded_div

    start = 0
    end = max_angle + bin_width * 0.01
    bin_edges = np.arange(start, end, bin_width)
    return bin_edges, bin_width


def _calc_locs(bin_width: float, axial: bool) -> np.ndarray:
    """
    Calculate bar plot bar locations.

    E.g.

    >>> _calc_locs(15, True)
    array([  7.5,  22.5,  37.5,  52.5,  67.5,  82.5,  97.5, 112.5, 127.5,
           142.5, 157.5, 172.5])

    :param bin_width: Real bin width
    :type bin_width: float
    :return: Array of locations
    """
    max_angle = 180 if axial else 360
    start = bin_width / 2
    end = max_angle + bin_width / 2
    locs = np.arange(start, end, bin_width)
    assert isinstance(locs, np.ndarray)
    return locs


def determine_azimuth_bins(
    azimuth_array: np.ndarray,
    length_array: Optional[np.ndarray] = None,
    bin_multiplier: Number = 1,
    axial: bool = True,
) -> AzimuthBins:
    """
    Calculate azimuth bins for plotting azimuth rose plots.

    E.g.

    >>> azimuth_array = np.array([25, 50, 145, 160])
    >>> length_array = np.array([5, 5, 10, 60])
    >>> azi_bins = determine_azimuth_bins(azimuth_array, length_array)
    >>> azi_bins.bin_heights
    array([ 5,  5,  0, 70])
    >>> azi_bins.bin_locs
    array([ 22.5,  67.5, 112.5, 157.5])
    >>> azi_bins.bin_width
    45.0

    """
    # Ideal width of rose plot bin based on sample size.
    # Lower bin size with bin_multiplier (result no longer ideal!)
    ideal_bin_width = _calc_ideal_bin_width(len(azimuth_array), axial=axial)
    ideal_bin_width = ideal_bin_width * bin_multiplier
    # True rose plot width.
    bin_edges, bin_width = _calc_bins(ideal_bin_width, axial=axial)
    # Location of rose plot bins.
    bin_locs = _calc_locs(bin_width, axial=axial)
    if length_array is None:
        # If no length_array is passed weight of 1.0 for all means equal
        # weights.
        length_array = np.array([1.0] * (len(azimuth_array)))
    # Height of rose plot bins.
    bin_heights, _ = np.histogram(azimuth_array, bin_edges, weights=length_array)

    return AzimuthBins(bin_width=bin_width, bin_locs=bin_locs, bin_heights=bin_heights)


def plot_azimuth_ax(
    bin_width: float,
    bin_locs: np.ndarray,
    bin_heights: np.ndarray,
    bar_color: str,
    ax: PolarAxes,
    axial: bool = True,
):
    """
    Plot azimuth rose plot to ax.
    """
    # Rose type always equal-area
    number_of_azimuths = np.sqrt(bin_heights)

    # Plot azimuth rose plot
    ax.bar(
        np.deg2rad(bin_locs),
        number_of_azimuths,
        width=np.deg2rad(bin_width),
        bottom=0.0,
        color=bar_color,
        edgecolor="k",
        alpha=0.85,
        zorder=4,
    )

    # Plot setup
    ax.set_theta_zero_location("N")
    ax.set_theta_direction(-1)
    max_theta = 180 if axial else 360
    ax.set_thetagrids(
        np.arange(0, max_theta + 1, 45),
        fontweight="bold",
        fontfamily="DejaVu Sans",
        fontsize=11,
        alpha=0.95,
    )
    ax.set_thetamin(0)
    ax.set_thetamax(max_theta)
    # The average of number_of_azimuths is displayed as a radial grid-line.
    # TODO: Cannot modify theta lines or r lines
    _, _ = ax.set_rgrids(
        radii=(number_of_azimuths.mean(),),  # type: ignore
        angle=0,
        fontsize=1,
        alpha=0.8,
        fmt="",
        ha="left",
    )

    # Tick labels
    labels = ax.get_xticklabels()
    for label in labels:
        label._y = -0.03
        label._fontproperties._size = 15
        label._fontproperties._weight = "bold"
        if not axial and label == labels[-1]:
            # Remove tick at 360 degrees
            label.set(visible=False)
    return ax


def _create_azimuth_set_text(
    length_array: np.ndarray,
    set_array: np.ndarray,
    set_names: Tuple[str, ...],
    add_abundance_order: bool,
) -> str:
    """
    Create azimuth set statistics for figure.

    E.g.

    >>> length_array = np.array([5, 5, 10, 60])
    >>> set_array = np.array([str(val) for val in [1, 1, 2, 2]])
    >>> set_names = ["1", "2"]
    >>> print(_create_azimuth_set_text(length_array, set_array, set_names, False))
    1 = 12.5%
    2 = 87.5%

    """
    # Sum of all length of all sets
    sum_length = length_array.sum()

    # String to collect text in
    t = ""

    # Ratio of all length for each set
    ratios = [
        sum(length_array[set_array == set_name]) / sum_length for set_name in set_names
    ]

    # Sorted from largest ratio to smallest
    sorted_ratios = sorted(ratios, reverse=True)

    # Abundance order where 1 has the highest ratio
    abundance_order = [sorted_ratios.index(ratio) + 1 for ratio in ratios]

    for idx, (set_name, ratio, order) in enumerate(
        zip(set_names, ratios, abundance_order)
    ):
        text = f"{set_name} = {ratio:.1%}"
        if add_abundance_order:
            text += f" ({order}.)"
        if idx < len(set_names) - 1:
            text = text + "\n"
        t = t + text
    return t


def decorate_azimuth_ax(
    ax: PolarAxes,
    label: str,
    length_array: np.ndarray,
    set_array: np.ndarray,
    set_names: Tuple[str, ...],
    set_ranges: SetRangeTuple,
    axial: bool,
    visualize_sets: bool,
    append_azimuth_set_text: bool = False,
    add_abundance_order: bool = False,
):
    """
    Decorate azimuth rose plot ax.
    """
    # Title is the name of the target area or group
    # prop_title = dict(boxstyle="square", facecolor="linen", alpha=1, linewidth=2)
    # title = "\n".join(wrap(f"{label}", 10))
    title = fill(label, 10)
    ax.set_title(
        title,
        x=0.97 if axial else 1.15,
        y=0.8 if axial else 1.0,
        fontsize="xx-large",
        fontweight="bold",
        fontfamily="DejaVu Sans",
        va="top",
        # bbox=prop_title,
        transform=ax.transAxes,
        ha="center",
    )
    # prop = dict(boxstyle="square", facecolor="linen", alpha=1, pad=0.45)
    # text = f"n ={len(set_array)}\n"
    text = f"n = {len(set_array)}"
    if append_azimuth_set_text:
        text += "\n"
        text = text + _create_azimuth_set_text(
            length_array, set_array, set_names, add_abundance_order=add_abundance_order
        )
    ax.text(
        x=0.92 if axial else 1.1,
        y=0.03 if axial else 0.15,
        s=text,
        transform=ax.transAxes,
        fontsize="large",
        # weight="roman",
        # bbox=prop,
        fontfamily="DejaVu Sans",
        va="top",
        ha="center",
        ma="left",
    )

    # Add lines to denote azimuth set edges
    if visualize_sets:
        for set_range in set_ranges:
            for edge in set_range:
                ax.axvline(np.deg2rad(edge), linestyle="dashed", color="black")


def plot_azimuth_plot(
    azimuth_array: np.ndarray,
    length_array: np.ndarray,
    azimuth_set_array: np.ndarray,
    azimuth_set_names: Tuple[str, ...],
    azimuth_set_ranges: SetRangeTuple,
    label: str,
    plain: bool,
    append_azimuth_set_text: bool = False,
    add_abundance_order: bool = False,
    axial: bool = True,
    visualize_sets: bool = False,
    bar_color: str = "darkgrey",
) -> Tuple[AzimuthBins, Figure, PolarAxes]:
    """
    Plot azimuth rose plot to its own figure.

    Returns rose plot bin parameters, figure, ax
    """
    azimuth_bins = determine_azimuth_bins(azimuth_array, length_array, axial=axial)
    fig, ax = plt.subplots(subplot_kw=dict(polar=True), figsize=(6.5, 5.1))
    assert isinstance(ax, PolarAxes)
    assert isinstance(fig, Figure)
    ax = plot_azimuth_ax(
        bin_heights=azimuth_bins.bin_heights,
        bin_locs=azimuth_bins.bin_locs,
        bin_width=azimuth_bins.bin_width,
        ax=ax,
        axial=axial,
        bar_color=bar_color,
    )
    if not plain:
        decorate_azimuth_ax(
            ax=ax,
            label=label,
            length_array=length_array,
            set_array=azimuth_set_array,
            set_names=azimuth_set_names,
            set_ranges=azimuth_set_ranges,
            append_azimuth_set_text=append_azimuth_set_text,
            add_abundance_order=add_abundance_order,
            axial=axial,
            visualize_sets=visualize_sets,
        )
    return (
        azimuth_bins,
        fig,
        ax,
    )



"""
Contains main entrypoint class for validating trace data, Validation.

Create Validation objects from traces and their target areas to validate
the traces for further analysis (``fractopo.analysis.network.Network``).
"""
import logging
from dataclasses import dataclass
from itertools import chain
from textwrap import dedent
from typing import Any, List, Optional, Set, Tuple

import geopandas as gpd
from geopandas.sindex import PyGEOSSTRTreeIndex
from shapely.geometry import LineString, MultiLineString, Point

from fractopo.general import (
    check_for_z_coordinates,
    determine_general_nodes,
    is_empty_area,
    remove_z_coordinates_from_geodata,
)
from fractopo.tval import trace_validators
from fractopo.tval.trace_validation_utils import determine_trace_candidates
from fractopo.tval.trace_validators import (
    ALL_VALIDATORS,
    MAJOR_ERRORS,
    MAJOR_VALIDATORS,
    VALIDATION_REQUIRES_NODES,
    MultiJunctionValidator,
    ValidatorClass,
)

log = logging.getLogger(__name__)


@dataclass
class Validation:

    """
    Validate traces data delineated by target area(s).

    If allow_fix is True, some automatic fixing will be done to e.g. convert
    MultiLineStrings to LineStrings.
    """

    traces: gpd.GeoDataFrame
    area: gpd.GeoDataFrame
    name: str
    allow_fix: bool

    # Default thresholds
    SNAP_THRESHOLD: float = 0.01
    SNAP_THRESHOLD_ERROR_MULTIPLIER: float = 1.1
    AREA_EDGE_SNAP_MULTIPLIER: float = 1.5
    TRIANGLE_ERROR_SNAP_MULTIPLIER: float = 10.0
    OVERLAP_DETECTION_MULTIPLIER: float = 50.0
    # TODO: Separate SHARP TURNS to major and minor
    SHARP_AVG_THRESHOLD: float = 135.0
    SHARP_PREV_SEG_THRESHOLD: float = 100.0
    ERROR_COLUMN: str = "VALIDATION_ERRORS"
    GEOMETRY_COLUMN: str = "geometry"

    determine_validation_nodes: bool = True

    def __post_init__(self):
        """
        Initialize private attributes used for caching.
        """
        # Private caching attributes
        self._endpoint_nodes: Optional[List[Tuple[Point, ...]]] = None
        self._intersect_nodes: Optional[List[Tuple[Point, ...]]] = None
        self._spatial_index: Optional[PyGEOSSTRTreeIndex] = None
        self._faulty_junctions: Optional[Set[int]] = None
        self._vnodes: Optional[Set[int]] = None

        self.ERROR_COLUMN_TRUNC = (
            self.ERROR_COLUMN[0:10]
            if len(self.ERROR_COLUMN) > 9
            else self.ERROR_COLUMN[0 : len(self.ERROR_COLUMN)]
        )

        if check_for_z_coordinates(geodata=self.traces):
            log.warning(
                dedent(
                    """
                Traces inputted into Validation contain Z-coordinates. They
                will be removed from the input traces and subsuquently from all
                outputs.
                """
                ).strip()
            )
            traces_with_removed_z = remove_z_coordinates_from_geodata(
                geodata=self.traces
            )
            assert isinstance(traces_with_removed_z, gpd.GeoDataFrame)
            self.traces = traces_with_removed_z

    def set_general_nodes(self):
        """
        Set _intersect_nodes and _endpoint_nodes attributes.
        """
        self._intersect_nodes, self._endpoint_nodes = determine_general_nodes(
            self.traces
        )

    @property
    def endpoint_nodes(self) -> List[Tuple[Point, ...]]:
        """
        Get endpoints of all traces.

        Returned as a list of tuples wherein each tuple represents the nodes
        of a trace in traces i.e. endpoint_nodes[index] are the nodes for
        traces[index].
        """
        if self._endpoint_nodes is None:
            self.set_general_nodes()
        if self._endpoint_nodes is not None:
            return self._endpoint_nodes
        raise TypeError("Expected self._endpoint_nodes to not be None.")

    @property
    def intersect_nodes(self) -> List[Tuple[Point, ...]]:
        """
        Get intersection nodes of all traces.

        Returned as a list of tuples wherein each tuple represents the nodes
        of a trace in traces i.e. intersect_nodes[index] are the nodes for
        traces[index].
        """
        if self._intersect_nodes is None:
            self.set_general_nodes()
        if self._intersect_nodes is not None:
            return self._intersect_nodes
        raise TypeError("Expected self._intersect_nodes to not be None.")

    @property
    def spatial_index(self) -> Optional[PyGEOSSTRTreeIndex]:
        """
        Get geopandas/pygeos spatial_index of traces.
        """
        if self._spatial_index is None:
            spatial_index = self.traces.sindex
            if (
                not isinstance(spatial_index, PyGEOSSTRTreeIndex)
                or len(spatial_index) == 0
            ):
                log.warning(
                    "Expected sindex property to be of type: PyGEOSSTRTreeIndex \n"
                    "and non-empty."
                )
                self._spatial_index = None
                return self._spatial_index
            self._spatial_index = spatial_index

        return self._spatial_index

    @property
    def faulty_junctions(self) -> Optional[Set[int]]:
        """
        Determine indexes with Multi Junctions.
        """
        if self._faulty_junctions is None and self.determine_validation_nodes:
            all_nodes = [
                tuple(chain(first, second))
                for first, second in zip(self.intersect_nodes, self.endpoint_nodes)
            ]
            self._faulty_junctions = MultiJunctionValidator.determine_faulty_junctions(
                all_nodes,
                snap_threshold=self.SNAP_THRESHOLD,
                snap_threshold_error_multiplier=self.SNAP_THRESHOLD_ERROR_MULTIPLIER,
            )
        return self._faulty_junctions

    @property
    def vnodes(self) -> Optional[Set[int]]:
        """
        Determine indexes with V-Nodes.
        """
        if self._vnodes is None and self.determine_validation_nodes:
            self._vnodes = trace_validators.VNodeValidator.determine_v_nodes(
                endpoint_nodes=self.endpoint_nodes,
                snap_threshold=self.SNAP_THRESHOLD,
                snap_threshold_error_multiplier=self.SNAP_THRESHOLD_ERROR_MULTIPLIER,
            )
        return self._vnodes

    def run_validation(
        self,
        first_pass=True,
        choose_validators: Optional[Tuple[ValidatorClass]] = None,
        allow_empty_area: bool = True,
    ) -> gpd.GeoDataFrame:
        """
        Run validation.

        Main entrypoint for validation. Returns validated traces GeoDataFrame.
        """
        for err_col in (self.ERROR_COLUMN, self.ERROR_COLUMN_TRUNC):
            if err_col in self.traces.columns:
                log.info(
                    "Dropping existing validation columns from traces GeoDataFrame."
                )
                traces = self.traces.drop(columns=err_col)
                assert isinstance(traces, gpd.GeoDataFrame)
                self.traces = traces

        # There's an option to choose the validators
        # Otherwise major are selected when running the first validation pass
        # Otherwise all are selected
        validators = MAJOR_VALIDATORS if first_pass else ALL_VALIDATORS
        if choose_validators is not None:
            validators = choose_validators

        # Check if chosen validators require determining nodes.
        self.determine_validation_nodes = any(
            validator in VALIDATION_REQUIRES_NODES for validator in validators
        )

        # Check if target area is completely void of traces
        if not allow_empty_area and is_empty_area(area=self.area, traces=self.traces):
            log.error(f"No traces within target area with name: {self.name}.")
            empty_gdf: gpd.GeoDataFrame = self.traces.copy()
            return empty_gdf

        all_errors: List[List[str]] = []
        all_geoms: List[LineString] = []
        for idx, geom in enumerate(self.traces.geometry.values):
            # Collect errors from each validator for each geom
            current_errors: List[str] = []

            # If geom contains validation error that will cause issues in later
            # validation -> ignore the geom and break out of validation loop
            # for current geom.
            ignore_geom: bool = False
            trace_candidates: Optional[gpd.GeoSeries] = None

            # validation loop
            for validator in validators:
                if ignore_geom:
                    # Break out of validation loop. See above comments
                    break

                if isinstance(geom, LineString) and not geom.is_empty:
                    # Some conditionals to avoid try-except loop
                    # trace candidates that are nearby to geom based on spatial index
                    trace_candidates = (
                        determine_trace_candidates(
                            geom, idx, self.traces, spatial_index=self.spatial_index
                        )
                        if trace_candidates is None
                        else trace_candidates
                    )

                # Overwrites geom if fix was executed
                # current_errors either contains new error or is unchanged
                # ignore_geom overwritten with True when geom must be ignored.
                geom, current_errors, ignore_geom = self._validate(
                    geom=geom,
                    validator=validator,
                    current_errors=current_errors,
                    allow_fix=self.allow_fix,
                    idx=idx,
                    snap_threshold=self.SNAP_THRESHOLD,
                    snap_threshold_error_multiplier=self.SNAP_THRESHOLD_ERROR_MULTIPLIER,
                    overlap_detection_multiplier=self.OVERLAP_DETECTION_MULTIPLIER,
                    triangle_error_snap_multiplier=self.TRIANGLE_ERROR_SNAP_MULTIPLIER,
                    trace_candidates=trace_candidates,
                    sharp_avg_threshold=self.SHARP_AVG_THRESHOLD,
                    sharp_prev_seg_threshold=self.SHARP_PREV_SEG_THRESHOLD,
                    area=self.area,
                    area_edge_snap_multiplier=self.AREA_EDGE_SNAP_MULTIPLIER,
                    spatial_index=self.spatial_index,
                    vnodes=self.vnodes,
                    faulty_junctions=self.faulty_junctions,
                )
            all_errors.append(current_errors)
            all_geoms.append(geom)

        assert len(all_errors) == len(all_geoms)
        validated_gdf = self.traces.copy()
        validated_gdf[self.ERROR_COLUMN] = all_errors
        validated_gdf[self.GEOMETRY_COLUMN] = all_geoms
        if first_pass:
            self.traces = validated_gdf
            # Run validation again
            validated_gdf = self.run_validation(
                first_pass=False, choose_validators=choose_validators
            )

        # Convert ERROR_COLUMN value types to tuples for better
        # interoperability with e.g. transformation to GeoJSON
        validated_gdf[self.ERROR_COLUMN] = [
            tuple(value) for value in validated_gdf[self.ERROR_COLUMN].values
        ]
        return validated_gdf

    @staticmethod
    def _validate(
        geom: Any,
        validator: ValidatorClass,
        current_errors: List[str],
        allow_fix: bool,
        **kwargs,
    ) -> Tuple[Any, List[str], bool]:
        """
        Validate geom with validator.

        Returns possibly fixed geom (if allow_fix is True and validator handles
        fixing), updated current_errors list for geom and whether to ignore
        the geom in later validations (e.g. when MultiLineString could not
        be merged by GeomTypeValidator fix.).

        Some validators require many additional kwargs.

        >>> geom = MultiLineString([((0, 0), (1, 1)), ((1, 1), (2, 2))])
        >>> validator = trace_validators.GeomTypeValidator
        >>> current_errors = []
        >>> allow_fix = True
        >>> fixed_geom, updated_errors, ignore_geom = Validation._validate(
        ...     geom=geom,
        ...     validator=validator,
        ...     current_errors=current_errors,
        ...     allow_fix=allow_fix,
        ... )
        >>> fixed_geom.wkt, updated_errors, ignore_geom
        ('LINESTRING (0 0, 1 1, 2 2)', [], False)

        """
        ignore_geom = False
        fixed = None
        if validator.LINESTRING_ONLY and not isinstance(geom, LineString):
            # Do not pass invalid geometry types to most validators. There's
            # already a error string in current_errors for e.g. MultiLineString
            # or empty geom rows.
            if isinstance(geom, MultiLineString):
                log.debug("MultiLineString geometry with validator ls only.")
            return geom, current_errors, True
        if (
            not validator.validation_method(geom=geom, **kwargs)
            and validator.ERROR not in current_errors
        ):
            # geom is invalid
            current_errors.append(validator.ERROR)
            if allow_fix:
                # Try to fix it
                try:
                    # fixed is None if fix is not successful but no error
                    # is raised
                    fixed = validator.fix_method(geom=geom, **kwargs)
                except NotImplementedError:
                    # No fix implemented for validator
                    fixed = None
                if fixed is not None:
                    # Fix successful
                    current_errors.remove(validator.ERROR)
                    # geom is passed to later validators
                    geom = fixed
            if validator.ERROR in MAJOR_ERRORS and fixed is None:
                # If error was not fixed and its part of major errors ->
                # the geom must be ignore by other validators.
                ignore_geom = True

        return geom, current_errors, ignore_geom



"""
fractopo.

Fracture Network Analysis
"""
import logging
import os

# Make GeoPandas use shapely 2.0 instead of pygeos
os.environ["USE_PYGEOS"] = "0"


from fractopo.analysis.multi_network import MultiNetwork  # noqa: E402,C0413
from fractopo.analysis.network import Network  # noqa: E402,C0413
from fractopo.tval.trace_validation import Validation  # noqa: E402,C0413

log = logging.getLogger(__name__)

__version__ = "0.6.0"


log.info(
    "Main imports available from fractopo/__init__.py:"
    f" {Network, Validation, MultiNetwork}"
)



"""
Anisotropy of connectivity determination utilities.
"""
import logging
from typing import Tuple

import matplotlib
import numpy as np
from matplotlib import patches
from matplotlib import pyplot as plt
from scipy.interpolate import CubicSpline

log = logging.getLogger(__name__)


def determine_anisotropy_classification(branch_classification: str) -> int:
    """
    Return value based on branch classification.

    Only C-C branches have a value, but this can be changed here.
    Classification can differ from 'C - C', 'C - I', 'I - I' (e.g. 'C - E') in
    which case a value (0) is still returned.

    :param branch_classification: Branch classification string.
    :return: Classification encoded as integer.

    E.g.

    >>> determine_anisotropy_classification("C - C")
    1

    >>> determine_anisotropy_classification("C - E")
    0
    """
    if branch_classification not in ("C - C", "C - I", "I - I"):
        return 0
    if branch_classification == "C - C":
        return 1
    if branch_classification == "C - I":
        return 0
    if branch_classification == "I - I":
        return 0
    return 0


def determine_anisotropy_sum(
    azimuth_array: np.ndarray,
    branch_types: np.ndarray,
    length_array: np.ndarray,
    sample_intervals: np.ndarray = np.arange(0, 179, 30),
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Determine the sums of branch anisotropies.

    :param azimuth_array: Array of branch azimuth values.
    :param branch_types: Array of branch type classication strings.
    :param length_array: Array of branch lengths.
    :param sample_intervals: Array of the sampling intervals.
    :return: Sums of branch anisotropies.

    E.g.

    >>> from pprint import pprint
    >>> azimuths = np.array([20, 50, 60, 70])
    >>> lengths = np.array([2, 5, 6, 7])
    >>> branch_types = np.array(["C - C", "C - C", "C - C", "C - I"])
    >>> pprint(determine_anisotropy_sum(azimuths, branch_types, lengths))
    (array([ 8.09332329, 11.86423103, 12.45612765,  9.71041492,  5.05739707,
            2.15381611]),
     array([  0,  30,  60,  90, 120, 150]))

    """
    # Array of anisotropy_array-s
    anisotropy_arrays = np.array(
        [
            determine_anisotropy_value(azimuth, branch_type, length, sample_intervals)
            for azimuth, branch_type, length in zip(
                azimuth_array, branch_types, length_array
            )
        ]
    )
    return anisotropy_arrays.sum(axis=0), sample_intervals


def determine_anisotropy_value(
    azimuth: float,
    branch_type: str,
    length: float,
    sample_intervals: np.ndarray = np.arange(0, 179, 30),
) -> np.ndarray:
    """
    Calculate anisotropy of connectivity for a branch.

    Based on azimuth, branch_type and length.  Value is calculated for preset
    angles (sample_intervals = np.arange(0, 179, 30))

    E.g.

    Anisotropy for a C-C classified branch:

    >>> determine_anisotropy_value(50, "C - C", 1)
    array([0.64278761, 0.93969262, 0.98480775, 0.76604444, 0.34202014,
           0.17364818])

    Other classification for branch:

    >>> determine_anisotropy_value(50, "C - I", 1)
    array([0, 0, 0, 0, 0, 0])

    """
    classification = determine_anisotropy_classification(branch_type)
    # CALCULATION
    results = []
    for angle in sample_intervals:
        if classification == 0:
            results.append(0)
            continue
        diff = np.abs(angle - azimuth)
        if diff > 90:
            diff = 180 - max([angle, azimuth]) + min([angle, azimuth])
        cos_diff = np.cos(np.deg2rad(diff))  # type: ignore
        result = length * classification * cos_diff
        results.append(result)
    # print(results)
    return np.array(results)


def plot_anisotropy_plot(
    anisotropy_sum: np.ndarray,
    sample_intervals: np.ndarray,
    # label: str,
    # color: Optional[str] = None,
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:  # type: ignore
    """
    Plot anisotropy values to new figure.
    """
    fig, ax = plt.subplots(subplot_kw=dict(polar=True))
    plot_anisotropy_ax(
        anisotropy_sum=anisotropy_sum, ax=ax, sample_intervals=sample_intervals
    )
    return fig, ax


def plot_anisotropy_ax(
    anisotropy_sum: np.ndarray,
    ax: plt.PolarAxes,
    sample_intervals: np.ndarray = np.arange(0, 179, 30),
):
    """
    Plot a styled anisotropy of connectivity to a given PolarAxes.

    Spline done with:
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html
    """
    double_anisotropy = np.concatenate([anisotropy_sum, anisotropy_sum])
    opp_angles = [i + 180 for i in sample_intervals]
    angles = list(sample_intervals) + opp_angles
    # PLOT SETUP
    ax.set_theta_zero_location("N")
    ax.set_theta_direction(-1)
    max_aniso = max(anisotropy_sum)

    for theta, r in zip(angles, double_anisotropy):
        theta = np.deg2rad(theta)  # type: ignore
        arrowstyle = patches.ArrowStyle.CurveB(head_length=1, head_width=0.5)
        ax.annotate(
            "",
            xy=(theta, r),
            xytext=(theta, 0),
            arrowprops=dict(
                edgecolor="black", facecolor="seashell", arrowstyle=arrowstyle
            ),
        )
    ax.axis("off")
    # CREATE CURVED STRUCTURE AROUND SCATTER AND ARROWS
    angles.append(359.999)
    double_anisotropy = np.concatenate([double_anisotropy, double_anisotropy[0:1]])
    angles_arr = np.array(angles)

    # https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.CubicSpline.html
    theta = np.deg2rad(angles_arr)  # type: ignore
    cs = CubicSpline(theta, double_anisotropy, bc_type="periodic")
    xnew = np.linspace(theta.min(), theta.max(), 300)
    power_smooth = cs(xnew)
    ax.plot(xnew, power_smooth, linewidth=1.5, color="black")

    circle = patches.Circle(
        (0, 0),
        0.0025 * max_aniso,
        transform=ax.transData._b,
        edgecolor="black",
        facecolor="gray",
        zorder=10,
    )
    ax.add_artist(circle)



"""
Command-line integration of fractopo with click.
"""
import json
import logging
import sys
import time
from enum import Enum, unique
from itertools import chain
from pathlib import Path
from typing import Dict, Optional, Tuple, Type

import click
import fiona
import geopandas as gpd
import pandas as pd
import pygeos
import typer
from rich.console import Console
from rich.table import Table
from rich.text import Text
from typer import Typer

from fractopo import __version__
from fractopo.analysis.network import Network
from fractopo.general import check_for_wrong_geometries, read_geofile, save_fig
from fractopo.tval.trace_validation import Validation
from fractopo.tval.trace_validators import SharpCornerValidator, TargetAreaSnapValidator

log = logging.getLogger(__name__)

# Initialize typer command-line app object
APP = Typer()

# Use minimum console width of 80
CONSOLE = Console(width=min([80, Console().width]))

SNAP_THRESHOLD_HELP = (
    "Distance threshold used to estimate whether e.g. a trace abuts another."
)
TRACE_FILE_HELP = "Path to lineament or fracture trace data."
AREA_FILE_HELP = "Path to target area data that delineates trace data."


@unique
class LogLevel(Enum):

    """
    Enums for log levels.
    """

    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def get_click_path_args(exists=True, **kwargs):
    """
    Get basic click path args.
    """
    path_arguments = dict(
        type=click.Path(
            exists=exists, file_okay=True, dir_okay=False, resolve_path=True, **kwargs
        ),
        nargs=1,
    )
    return path_arguments


def _check_for_wrong_geometries_cli(traces: gpd.GeoDataFrame, area: gpd.GeoDataFrame):
    try:
        check_for_wrong_geometries(traces=traces, area=area)
    except TypeError:
        CONSOLE.print(
            Text.assemble(
                "Check that traces and target area arguments are passed in"
                " the correct order in the command-line (Check with --help)."
            )
        )
        raise


def describe_results(
    validated: gpd.GeoDataFrame, error_column: str, console: Console = CONSOLE
):
    """
    Describe validation results to stdout.
    """
    error_count = sum(len(val) != 0 for val in validated[error_column].values)
    error_types = {
        c for c in chain(*validated[error_column].to_list()) if isinstance(c, str)
    }
    trace_count = validated.shape[0]
    count_string = f"Out of {trace_count} traces, {error_count} were invalid."
    type_string = f"There were {len(error_types)} error types. These were:\n"
    type_color = "yellow"
    for error_type in error_types:
        type_string += error_type + "\n"
        if SharpCornerValidator.ERROR not in error_type:
            type_color = "bold red"
    count_color = "bold red" if error_count / trace_count > 0.05 else "yellow"
    console.print(Text.assemble((count_string, count_color)))
    if len(error_types) > 0:
        console.print(Text.assemble((type_string, type_color)))


def make_output_dir(base_path: Path) -> Path:
    """
    Make timestamped output dir.
    """
    localtime = time.localtime()
    tm_min = localtime.tm_min
    hour = localtime.tm_hour
    day = localtime.tm_mday
    month = localtime.tm_mon
    year = localtime.tm_year
    timestr = "_".join(map(str, [day, month, year, hour, tm_min]))
    output_dir = base_path / f"validated_{timestr}"
    if not output_dir.exists():
        output_dir.mkdir()
    return output_dir


def rich_table_from_parameters(parameters: Dict[str, float]) -> Table:
    """
    Generate ``rich`` ``Table`` from network parameters.
    """
    param_table = Table(
        title="Selected Network Parameters",
    )
    param_table.add_column("Parameter", header_style="bold", style="bold green")
    param_table.add_column("Value", header_style="bold", style="blue")

    for key, value in parameters.items():
        param_table.add_row(key, f"{value:.4f}")
    return param_table


@APP.callback()
def fractopo_callback(
    log_level: LogLevel = typer.Option(LogLevel.WARNING.value),
):
    """
    Use fractopo command-line utilities.
    """
    log_level_int = int(getattr(logging, log_level.value))
    log.info("Setting up log with basicConfig.")
    logging.basicConfig(level=log_level_int, force=True)


@APP.command()
def tracevalidate(
    trace_file: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
        help=TRACE_FILE_HELP,
    ),
    area_file: Path = typer.Argument(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
        help=AREA_FILE_HELP,
    ),
    allow_fix: bool = typer.Option(
        True,
        "--allow-fix",
        "--fix",
        help=(
            "Enable the direct modification of output trace file to fix errors. "
            "Input files will not be modified unless specified as the --output target."
        ),
    ),
    summary: bool = typer.Option(True, help="Print summary of validation results."),
    snap_threshold: float = typer.Option(
        0.001,
        help="Distance threshold used to estimate whether e.g. a trace abuts another.",
    ),
    output: Optional[Path] = typer.Option(
        None, help="Where to save validated output trace data."
    ),
    only_area_validation: bool = typer.Option(
        False, help="Only validate the area boundary snapping."
    ),
    allow_empty_area: bool = typer.Option(
        True, help="Allow empty areas to validation."
    ),
):
    """
    Validate trace data delineated by target area data.

    If allow_fix is True, some automatic fixing will be done to e.g. convert
    MultiLineStrings to LineStrings.
    """
    # Assert that read files result in GeoDataFrames
    traces: gpd.GeoDataFrame = read_geofile(trace_file)
    areas: gpd.GeoDataFrame = read_geofile(area_file)
    if not all(isinstance(val, gpd.GeoDataFrame) for val in (traces, areas)):
        raise TypeError(
            "Expected trace and area files to be readable as GeoDataFrames."
        )

    _check_for_wrong_geometries_cli(traces=traces, area=areas)
    log.info(f"Validating traces: {trace_file} area: {area_file}.")
    # Get input crs
    input_crs = traces.crs

    # Validate
    validation = Validation(
        traces,
        areas,
        trace_file.stem,
        allow_fix,
        SNAP_THRESHOLD=snap_threshold,
    )
    if only_area_validation:
        CONSOLE.print(Text.assemble(("Only performing area validation.", "yellow")))
        choose_validators: Optional[Tuple[Type[TargetAreaSnapValidator]]] = (
            TargetAreaSnapValidator,
        )
    else:
        choose_validators = None
    CONSOLE.print(
        Text.assemble("Performing validation of ", (trace_file.name, "blue"), ".")
    )
    validated_trace = validation.run_validation(
        choose_validators=choose_validators, allow_empty_area=allow_empty_area
    )

    # Set same crs as input if input had crs
    if input_crs is not None:
        validated_trace.crs = input_crs

    # Get input driver to use as save driver
    with fiona.open(trace_file) as open_trace_file:
        assert open_trace_file is not None
        save_driver = open_trace_file.driver

    # Resolve output if not explicitly given
    if output is None:
        output_dir = make_output_dir(Path(".")).resolve()
        output_path = output_dir / f"{trace_file.stem}_validated{trace_file.suffix}"
        CONSOLE.print(
            Text.assemble(
                (
                    f"Generated output directory at {output_dir} "
                    f"where validated output will be saved in file {output_path.name}.",
                    "blue",
                )
            )
        )
    else:
        output_path = output

    # Remove file if one exists at output_path
    if output_path.exists():
        CONSOLE.print(
            Text.assemble(("Overwriting old file at given output path.", "yellow"))
        )
        output_path.unlink()

    # Change validation_error column to type: str and consequently save
    # the GeoDataFrame.
    assert not isinstance(validated_trace[validation.ERROR_COLUMN].iloc[0], list)
    validated_trace.astype({validation.ERROR_COLUMN: str}).to_file(
        output_path, driver=save_driver
    )
    if summary:
        describe_results(validated_trace, validation.ERROR_COLUMN, console=CONSOLE)


@APP.command()
def network(
    trace_file: Path = typer.Argument(
        ..., exists=True, dir_okay=False, help=TRACE_FILE_HELP
    ),
    area_file: Path = typer.Argument(
        ..., exists=True, dir_okay=False, help=AREA_FILE_HELP
    ),
    snap_threshold: float = typer.Option(0.001, help=SNAP_THRESHOLD_HELP),
    determine_branches_nodes: bool = typer.Option(
        True,
        help="Whether to determine branches and nodes as part of analysis. Recommended.",
    ),
    name: Optional[str] = typer.Option(
        None, help="Name for Network. Used when saving outputs and as plot titles."
    ),
    circular_target_area: bool = typer.Option(
        False, help="Is/are target area(s) circles?"
    ),
    truncate_traces: bool = typer.Option(
        True, help="Whether to cut traces at target area boundary. Recommended."
    ),
    censoring_area: Optional[Path] = typer.Option(
        None,
        help="Path to area data that delineates censored areas within target areas.",
    ),
    branches_output: Optional[Path] = typer.Option(
        None, help="Where to save branch data."
    ),
    nodes_output: Optional[Path] = typer.Option(None, help="Where to save node data."),
    general_output: Optional[Path] = typer.Option(
        None, help="Where to save general network analysis outputs e.g. plots."
    ),
    parameters_output: Optional[Path] = typer.Option(
        None, help="Where to save numerical parameter data from analysis."
    ),
):
    """
    Analyze the geometry and topology of trace network.
    """
    network_name = name if name is not None else area_file.stem

    CONSOLE.print(
        Text.assemble(
            "Performing network analysis of ", (network_name, "bold green"), "."
        )
    )
    traces = gpd.read_file(trace_file)
    areas = gpd.read_file(area_file)
    _check_for_wrong_geometries_cli(traces=traces, area=areas)

    network = Network(
        trace_gdf=traces,
        area_gdf=areas,
        snap_threshold=snap_threshold,
        determine_branches_nodes=determine_branches_nodes,
        name=network_name,
        circular_target_area=circular_target_area,
        truncate_traces=truncate_traces,
        censoring_area=read_geofile(censoring_area)
        if censoring_area is not None
        else gpd.GeoDataFrame(),
    )
    (
        general_output_path,
        branches_output_path,
        nodes_output_path,
        parameters_output_path,
    ) = default_network_output_paths(
        network_name=network_name,
        general_output=general_output,
        branches_output=branches_output,
        nodes_output=nodes_output,
        parameters_output=parameters_output,
    )
    if determine_branches_nodes:
        # Save branches and nodes
        CONSOLE.print(
            Text.assemble(
                "Saving branches to ",
                (str(branches_output_path), "bold green"),
                " and nodes to ",
                (str(nodes_output_path), "bold green"),
                ".",
            )
        )
        network.branch_gdf.to_file(branches_output_path, driver="GPKG")
        network.node_gdf.to_file(nodes_output_path, driver="GPKG")

    base_parameters = network.parameters

    # Print pretty table of basic network parameters
    CONSOLE.print(rich_table_from_parameters(base_parameters))

    parameters = (
        network.numerical_network_description()
        if determine_branches_nodes
        else base_parameters
    )

    pd.DataFrame([parameters]).to_csv(parameters_output_path)

    CONSOLE.print(
        Text.assemble(
            "Saving network parameter csv to path:\n",
            (str(parameters_output_path), "bold green"),
        )
    )

    if determine_branches_nodes:
        # Plot ternary XYI-node proportion plot
        fig, _, _ = network.plot_xyi()
        save_fig(fig=fig, results_dir=general_output_path, name="xyi_ternary_plot")

        # Plot ternary branch proportion plot
        fig, _, _ = network.plot_branch()
        save_fig(fig=fig, results_dir=general_output_path, name="branch_ternary_plot")

    # Plot trace azimuth rose plot
    _, fig, _ = network.plot_trace_azimuth()
    save_fig(fig=fig, results_dir=general_output_path, name="trace_azimuth")

    # Plot trace length distribution plot
    _, fig, _ = network.plot_trace_lengths()
    save_fig(fig=fig, results_dir=general_output_path, name="trace_length_distribution")

    # Save additional numerical data for possible post-processing needs in a
    # json file
    json_data_arrays = {
        "trace_length_array": network.trace_length_array,
        "branch_length_array": network.branch_length_array
        if determine_branches_nodes
        else None,
        "trace_azimuth_array": network.trace_azimuth_array,
        "branch_azimuth_array": network.branch_azimuth_array
        if determine_branches_nodes
        else None,
        "trace_azimuth_set_array": network.trace_azimuth_set_array,
        "branch_azimuth_set_array": network.branch_azimuth_set_array
        if determine_branches_nodes
        else None,
    }
    json_data_lists = {
        key: (item.tolist() if item is not None else None)
        for key, item in json_data_arrays.items()
    }
    json_data_path = general_output_path / "additional_numerical_data.json"

    json_data_path.write_text(json.dumps(json_data_lists, sort_keys=True))


def default_network_output_paths(
    network_name: str,
    general_output: Optional[Path],
    branches_output: Optional[Path],
    nodes_output: Optional[Path],
    parameters_output: Optional[Path],
):
    """
    Determine default network output paths.
    """
    general_output_path = (
        Path(f"{network_name}_outputs") if general_output is None else general_output
    )
    general_output_path.mkdir(exist_ok=True)

    branches_output_path = (
        general_output_path / f"{network_name}_branches.gpkg"
        if branches_output is None
        else branches_output
    )
    nodes_output_path = (
        general_output_path / f"{network_name}_nodes.gpkg"
        if nodes_output is None
        else nodes_output
    )
    parameters_output_path = (
        general_output_path / f"{network_name}_parameters.csv"
        if parameters_output is None
        else parameters_output
    )

    return (
        general_output_path,
        branches_output_path,
        nodes_output_path,
        parameters_output_path,
    )


@APP.command()
def info():
    """
    Print out information about fractopo installation and python environment.
    """
    information = dict(
        fractopo_version=__version__,
        geopandas_version=gpd.__version__,
        pygeos_version=pygeos.__version__,
        package_location=str(Path(__file__).parent.absolute()),
        python_location=str(Path(sys.executable).absolute()),
    )
    CONSOLE.print_json(json.dumps(information, sort_keys=True))
